/* FILE: dump.c
 *
 * Copyright (c) 2004-2006 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "config.h"

#define OSIVX_WITH_KEYLIST
#include <osivx.h>
#include <zargs.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>


typedef struct
{
  char* input_fn;
  char* output_fn;
  int   print_args;
  int   print_vecs;
  int   print_maps;
  long  frame_range[3];
  long  wind_num;
  char* vecs_fmt;
  char* corr_fmt; 
  int   omit_headrs;
  int   omit_nlines;
  int   omit_fbreak;
} dump_args;

typedef struct
{
  long frame_range[3];
} dump_info;

typedef struct
{
  unsigned*      mask;
  dump_args*  arglist;
  dump_info*  optinfo;
} dump_packet;

struct dump_fmtbox_t
{
   char* fmt;
   osivx_ioreal_t* addr;
   struct dump_fmtbox_t* next;
};

typedef struct dump_fmtbox_t dump_fmtbox;

enum {
  DUMP_OUTPUT_FN,   DUMP_PRINT_ARGS,
  DUMP_PRINT_VECS,  DUMP_PRINT_MAPS,
  DUMP_FRAME_NUM,   DUMP_WIND_NUM,
  DUMP_VECS_FMT,    DUMP_CORR_FMT,
  DUMP_OMIT_FBREAK, DUMP_OMIT_HEADRS,
  DUMP_OMIT_NLINES, DUMP_NOPTS
};

static const char* dump_usage =
"Usage: osiv_dump [ -o output ] [ ... ] FILE\n\n"
"More information is available in the 'osiv_dump' manpage as well\n"
"as in the online documentation.\n";

static unsigned dump_long_nopts = 5;
static zopts_item dump_long_opts[] = {
   {"arguments",     ZOPTS_NARG,  DUMP_PRINT_ARGS},
   {"maps",          ZOPTS_NARG,  DUMP_PRINT_MAPS},
   {"omit-fbreaks",  ZOPTS_NARG,  DUMP_OMIT_FBREAK},
   {"omit-headers",  ZOPTS_NARG,  DUMP_OMIT_HEADRS},
   {"omit-newlines", ZOPTS_NARG,  DUMP_OMIT_NLINES},
   {"vectors",       ZOPTS_NARG,  DUMP_PRINT_VECS},
   {NULL, 0, 0}
};

static unsigned dump_shrt_nopts = 8;
static zopts_item dump_shrt_opts[] = {
   {"a",             ZOPTS_NARG,  DUMP_PRINT_ARGS},
   {"c",             ZOPTS_WARG,  DUMP_CORR_FMT},
   {"f",             ZOPTS_WARG,  DUMP_FRAME_NUM},
   {"m",             ZOPTS_NARG,  DUMP_PRINT_MAPS},
   {"o",             ZOPTS_WARG,  DUMP_OUTPUT_FN},
   {"p",             ZOPTS_WARG,  DUMP_VECS_FMT},
   {"v",             ZOPTS_NARG,  DUMP_PRINT_VECS},
   {"w",             ZOPTS_WARG,  DUMP_WIND_NUM},
   {NULL, 0, 0}
};

#define ARGUMENT_MSG                                          \
"#  Arguments\n"                                              \
"########################################################\n"

#define VECTORS_MSG                                           \
"#  Vectors \n"                                               \
"########################################################\n"

#define FRAMES_MSG                                            \
"#  Frame: %ld\n" 

#define CORRELATION_MSG                                       \
"#  Correlation Map\n"                                        \
"########################################################\n"


enum {
  DUMP_EMULTD,         DUMP_ENINT,
  DUMP_EINTER,         DUMP_ESIGNAL,
  DUMP_ESIGHUP,        DUMP_ESIGINT,
  DUMP_ESIGQUIT,       DUMP_ESIGTERM,
  DUMP_ESIGFPE,        DUMP_ESIGILL,
  DUMP_ESIGSEGV,       DUMP_ESIGBUS,
  DUMP_ESIGABRT,       DUMP_EARGV,
  DUMP_EPNEED,         DUMP_EPWONT,
  DUMP_EPLOST,         DUMP_EPBUFF,
  DUMP_EMEM,           DUMP_EFOPEN,
  DUMP_EFILE,          DUMP_ESTDIN,
  DUMP_EFOUT,          DUMP_EBOTH,
  DUMP_EFORMAT,        DUMP_EMAPS,
  DUMP_EFRAME,         DUMP_ENRNG,
  DUMP_EFRANGE
};

#define PHONE_HOME \
"\nYou have most likely found a bug within the program.\n"\
"Your help in resolving this problem would be very much\n"\
"appreciated. Please visit the website at www.osiv.org,\n"\
"and submit a bug report which includes information on\n"\
"the situation (and file) with which this occured.\n"

static const char *
dump_strerr(
   int id)
{
   const char* s;

   switch(id)
   {
   case DUMP_EMULTD:
      s = "option %s defined more than once";
      break;
   case DUMP_ENINT:
      s = "option %s expects integer but found `%s\'";
      break;
   case DUMP_ESIGNAL:
      s = "unable to install signal handling";
      break;
   case DUMP_ESIGINT:
      s = "process caught user interupt";
      break;
   case DUMP_ESIGHUP:
      s = "process caught HUP signal";
      break;
   case DUMP_ESIGQUIT:
      s = "process caught QUIT signal";
      break;
   case DUMP_ESIGTERM:
      s = "process caught TERM signal";
      break;
   case DUMP_ESIGFPE:
      s = "caught floating point exception\n" PHONE_HOME;
      break;
   case DUMP_ESIGILL:
      s = "caught illegal instruction\n" PHONE_HOME;
      break;
   case DUMP_ESIGSEGV:
      s = "caught segmentation fault\n" PHONE_HOME;
      break;
   case DUMP_ESIGBUS:
      s = "caught bug error signal\n" PHONE_HOME;
      break;
   case DUMP_ESIGABRT:
      s = "caught runtime abort\n" PHONE_HOME;
      break;
   case DUMP_EARGV:
      s = "requires exactly one argument";
      break;
   case DUMP_EINTER:
      s = "caught internal program error\n" PHONE_HOME;
      break;
   case DUMP_EPNEED:
      s = "option `%s\' requires an argument";
      break;
   case DUMP_EPWONT:
      s = "option `%s\' doesn't take argument";
      break;
   case DUMP_EPLOST:
      s = "found unrecognized option `%s\'";
      break;
   case DUMP_EPBUFF:
      s = "argument to option `%s\' is too long";
      break;
   case DUMP_EMEM:
      s = "unable to allocate memory";
      break;
   case DUMP_EFOPEN:
      s = "unable to open the input file `%s\'";
      break;
   case DUMP_EFILE:
      s = "unable to parse input file";
      break;
   case DUMP_ESTDIN:
      s = "file cannot be read from standard input";
      break;
   case DUMP_EFOUT:
      s = "unable to open output file `%s\'";
      break;
   case DUMP_EBOTH:
      s = "cannot print both vectors and map";
      break;
   case DUMP_EFORMAT:
      s = "invalid print format string `%s'";
      break;
   case DUMP_EMAPS:
      s = "correlation maps not found";
      break;
   case DUMP_EFRAME:
      s = "no data for frame `%d' in file";
      break;
   case DUMP_ENRNG:
      s = "option %s expects range but found `%s\'";
      break;
   case DUMP_EFRANGE:
      s = "specified frame range is invalid";
      break;
   default:
      return NULL;
   } 

   return s;
}

static void
dump_die(
   int id,...)
{
   va_list ap;
   const char *c;

   fprintf(stderr,"osiv_dump error: ");
   c = dump_strerr(id);

   va_start(ap,id);
   vfprintf(stderr, c, ap);
   va_end(ap);

   fprintf(stderr, "\n");
   exit(1);
}

static void
dump_signal_dohard(
   int signum)
{
   switch (signum)
   {
   case SIGFPE:
      dump_die(DUMP_ESIGFPE);
      break;
   case SIGILL:
      dump_die(DUMP_ESIGILL);
      break;
   case SIGSEGV:
      dump_die(DUMP_ESIGSEGV);
      break;
   case SIGBUS:
      dump_die(DUMP_ESIGBUS);
      break;
   case SIGABRT:
      dump_die(DUMP_ESIGABRT);
      break;
   default:
      dump_die(DUMP_EINTER);
   }
}

static void
dump_signal_dosoft(
   int signum)
{
   switch (signum)
   {  
   case SIGHUP:
      dump_die(DUMP_ESIGHUP);
      break;
   case SIGINT:
      dump_die(DUMP_ESIGINT);
      break;
   case SIGQUIT:
      dump_die(DUMP_ESIGQUIT);
      break;
   case SIGTERM:
      dump_die(DUMP_ESIGTERM);
      break;
   default:
      dump_die(DUMP_EINTER);
   }
}

static void
dump_signal_mkhard()
{
   struct sigaction newact;
   sigset_t block_mask;

   /* make blocking mask */
   if (sigfillset(&block_mask))
      dump_die(DUMP_ESIGNAL);

   /* make action */
   memset(&newact,0,sizeof(newact));
   newact.sa_mask = block_mask;
   newact.sa_handler = &dump_signal_dohard;

   /* install handlers */
   if (sigaction(SIGFPE, &newact, NULL)  ||
       sigaction(SIGILL, &newact, NULL)  ||
       sigaction(SIGSEGV, &newact, NULL) ||
       sigaction(SIGBUS, &newact, NULL)  ||
       sigaction(SIGABRT, &newact, NULL))
     dump_die(DUMP_ESIGNAL);
}

static void
dump_signal_mksoft()
{
   struct sigaction newact;
   sigset_t block_mask;

   /* make blocking mask */
   if (sigfillset(&block_mask) ||
       sigdelset(&block_mask, SIGFPE)  ||
       sigdelset(&block_mask, SIGILL)  ||
       sigdelset(&block_mask, SIGSEGV) ||
       sigdelset(&block_mask, SIGBUS)  ||
       sigdelset(&block_mask, SIGABRT))
     dump_die(DUMP_ESIGNAL);

   /* make action */
   memset(&newact,0,sizeof(newact));
   newact.sa_mask = block_mask;
   newact.sa_handler = &dump_signal_dosoft;

   /* install handlers */
   if (sigaction(SIGHUP, &newact, NULL)  ||
       sigaction(SIGINT, &newact, NULL)  ||
       sigaction(SIGQUIT, &newact, NULL) ||
       sigaction(SIGTERM, &newact, NULL))
     dump_die(DUMP_ESIGNAL);
}

static void
dump_setint(
   unsigned* preset,
   const char* nm,
   const char* val,
   long* out)
{
   char* endptr = NULL;

   if (*preset)
     dump_die(DUMP_EMULTD,nm);

   *out = strtol(val,&endptr,10);
   if (*endptr != '\0' || val==endptr)
     dump_die(DUMP_ENINT,nm,val);

   *preset = 1; 
}

static void
dump_setbool(
   unsigned* preset,
   const char* nm,
   const char* val,
   int* out)
{
   if (*preset)
     dump_die(DUMP_EMULTD,nm);

   *out = 1;
   *preset = 1;
}

static void
dump_setstr(
   unsigned* preset,
   const char* nm,
   const char* val,
   char** out)
{
   if (*preset) 
     dump_die(DUMP_EMULTD,nm);

   *out = strdup(val);
   if (*out == NULL)
     dump_die(DUMP_EMEM);

   *preset = 1;
}

static void
dump_setrng(
   unsigned* preset,
   const char* nm,
   const char* val,
   long out[3])
{
   char* endptr = NULL;

   if (*preset)
     dump_die(DUMP_EMULTD,nm);

   out[0] = strtol(val,&endptr,10);
   if (*endptr != ':') {
      if (*endptr != '\0' || val==endptr)
        dump_die(DUMP_ENRNG,nm,val);

      out[1] = 1;
      out[2] = out[0];
   } else {
      const char* val2 = endptr + 1;
      out[1] = strtol(val2,&endptr,10);

      if (*endptr != ':') {
         if (*endptr != '\0' || val2==endptr)
           dump_die(DUMP_ENRNG,nm,val);

         out[2] = out[1];
         out[1] = -1;
      } else {
         const char* val3 = endptr + 1;
         out[2] = strtol(val3,&endptr,10);

         if (*endptr != '\0' || val3==endptr)
            dump_die(DUMP_ENRNG,nm,val);
      }
   }

   *preset = 1; 
}

static void
dump_optcb(
   long int evt,
   const char* var,
   const char* val,
   void* data)
{
   dump_packet* m =
     (dump_packet*)data;
   dump_args* a;

   a = m->arglist;
   switch(evt)
   {
   case DUMP_OUTPUT_FN:
      dump_setstr(&m->mask[evt],
        var, val, &a->output_fn);
      break;
   case DUMP_PRINT_ARGS:
      dump_setbool(&m->mask[evt],
        var, val, &a->print_args);
       break;
   case DUMP_PRINT_VECS:
      dump_setbool(&m->mask[evt],
        var, val, &a->print_vecs);
      break;
   case DUMP_PRINT_MAPS:
       dump_setbool(&m->mask[evt],
         var, val, &a->print_maps);
       break;
   case DUMP_FRAME_NUM:
      dump_setrng(&m->mask[evt],
        var, val, a->frame_range);
      break;
   case DUMP_WIND_NUM:
      dump_setint(&m->mask[evt],
        var, val, &a->wind_num);
      break;
   case DUMP_VECS_FMT: 
      dump_setstr(&m->mask[evt], 
        var, val, &a->vecs_fmt);
      break;
   case DUMP_CORR_FMT:
      dump_setstr(&m->mask[evt],
        var, val, &a->corr_fmt);
      break;
   case DUMP_OMIT_FBREAK:
      dump_setbool(&m->mask[evt],
        var, val, &a->omit_fbreak);
      break;
   case DUMP_OMIT_HEADRS:
      dump_setbool(&m->mask[evt],
        var, val, &a->omit_headrs);
      break;
   case DUMP_OMIT_NLINES:
      dump_setbool(&m->mask[evt],
        var, val, &a->omit_nlines);
      break;
   default: break;
   }

   switch(evt)
   {
   case ZOPTS_ENEED:
      dump_die(DUMP_EPNEED,var);
   case ZOPTS_EWONT:
      dump_die(DUMP_EPWONT,var);
   case ZOPTS_ELOST:
      dump_die(DUMP_EPLOST,var);
   case ZOPTS_EBUFF:
      dump_die(DUMP_EPBUFF,var);
   }
}


static void
dump_setinfo(
   dump_packet* m)
{
   dump_args* a = m->arglist;
   dump_info* p = m->optinfo;
   osivx_file_t* fi;
   long file_mark;
   osivx_ioint_t tmp0, tmp1, tmp2;

   /* open input file */
   fi = osivx_fopen(a->input_fn, "r");
   if (fi == NULL) {
     dump_die(DUMP_EFOPEN, a->input_fn);
   }

   /* collect frame info */
   if (osivx_fscan_totag (fi, OSIVX_TAG_HEADER) ||
       osivx_ftell (fi, &file_mark) ||
       osivx_fgrab_argint (fi, OSIVX_KEY_SFR, &tmp0) ||
       osivx_fseek (fi, file_mark, SEEK_SET) ||
       osivx_fgrab_argint (fi, OSIVX_KEY_SKP, &tmp1) ||
       osivx_fseek (fi, file_mark, SEEK_SET) ||
       osivx_fgrab_argint (fi, OSIVX_KEY_EFR, &tmp2) ||
       osivx_fseek(fi, 0, SEEK_SET)) {
      dump_die(DUMP_EFILE);
   }
   p->frame_range[0] = tmp0;
   p->frame_range[1] = tmp1;
   p->frame_range[2] = tmp2;

   /* close input file */
   osivx_fclose(fi);
}

 
static void
dump_setdft(
   dump_packet* m)
{
   unsigned* mask = m->mask;
   dump_args* a = m->arglist;
   dump_info* p = m->optinfo;

   if (!mask[DUMP_PRINT_ARGS] &&
       !mask[DUMP_PRINT_VECS] &&
       !mask[DUMP_PRINT_MAPS])
   {
      a->print_args = 1;
      a->print_vecs = 1;
      a->print_maps = 0;
   } else {
      if (!mask[DUMP_PRINT_ARGS]) {
         a->print_args = 0;
      }
      if (!mask[DUMP_PRINT_VECS]) {
         a->print_vecs = 0;
      }
      if (!mask[DUMP_PRINT_MAPS]) {
         a->print_maps = 0;
      }
   }

   if (!mask[DUMP_FRAME_NUM]) {
      a->frame_range[0] = p->frame_range[0];
      a->frame_range[1] = p->frame_range[1];
      a->frame_range[2] = p->frame_range[2];
   } else {
      if (a->frame_range[0] < 0) {
        a->frame_range[0] = p->frame_range[0];
      }
      if (a->frame_range[1] < 0) {
        a->frame_range[1] = p->frame_range[1];
      }
      if (a->frame_range[2] < 0) {
        a->frame_range[2] = p->frame_range[2];
      }
   }

   if (!mask[DUMP_WIND_NUM]) {
      a->wind_num = 0;
   }

   if (!mask[DUMP_VECS_FMT]) {
      a->vecs_fmt = strdup("%06.1x  %06.1y    %+0.8u  %+0.8v\\n");
   }
   if (!mask[DUMP_CORR_FMT]) {
      a->corr_fmt = strdup("%+05.1u  %+05.1v    %010.1e\\n");
   }

   if (!mask[DUMP_OMIT_HEADRS]) {
      a->omit_headrs = 0;
   }
   if (!mask[DUMP_OMIT_NLINES]) {
      a->omit_nlines = 0;
   }
   if (!mask[DUMP_OMIT_FBREAK]) {
      a->omit_fbreak = 0;
   }

   if (!mask[DUMP_OUTPUT_FN]) {
      a->output_fn = strdup("-");
   }
}

static void
dump_unescape (char *str)
{
   int i, nume = 8;
   char usc[] = { '\f', '\n', '\r', '\t', '\v', '\'', '\"', '\?' };
   char *esc[] = { "\\f", "\\n", "\\r", "\\t", "\\v", "\\\'", "\\\"", "\\\?" };

   for (i = 0; i < nume; i++)
   {
      char *c;
      while ((c = strstr (str, esc[i])) != NULL)
      {
         char *src, *dest;

         dest = c + 1;
         src = c + strlen (esc[i]);
         memmove (dest, src, strlen (src) + 1);

         *c = usc[i];
      }
   }
}

static char *
dump_arguments_convert (int key)
{
   char *t = (char *) malloc (OSIVX_STRLEN + 1);
   if (t == NULL) dump_die(DUMP_EMEM);

   if (key >= 0 && key <= OSIVX_KEY_MAX) {
      strcpy (t, osiv_keylist[key]);
   } else {
      sprintf (t, "unknown_variable_%d", key);
   }

   return t;
}

static void
dump_arguments_integer(
   osivx_file_t * fi,
   FILE * fo)
{
   char *vnm;
   osivx_ioint_t key, val;
   osivx_iotag_t ctag;

   /* read in value key and value */
   if (osivx_fread_argint (fi, &key, &val) ||
       osivx_fread_tag (fi, &ctag) ||
       osivx_tagcmp (ctag, OSIVX_GAT_ARGINT))
     dump_die(DUMP_EFILE);

   /* convert key to real */
   vnm = dump_arguments_convert (key);

   /* print out result */
   fprintf (fo, "%s = %d\n", vnm, val);

   /* free key */
   free (vnm);
}

static void
dump_arguments_string(
   osivx_file_t * fi,
   FILE * fo)
{
   char *val, *vnm;
   osivx_ioint_t key;
   osivx_iotag_t ctag;

   if (osivx_fread_argstr (fi, &key, &val) ||
       osivx_fread_tag (fi, &ctag) ||
       osivx_tagcmp (ctag, OSIVX_GAT_ARGSTR))
     dump_die(DUMP_EFILE);

   /* convert key */
   vnm = dump_arguments_convert (key);

   /* print out result */
   fprintf (fo, "%s = %s\n", vnm, val);

   /* free key */
   free (val);
   free (vnm);
}

static void
dump_arguments(
   osivx_file_t* fi,
   FILE* fo,
   int omit_headrs,
   int omit_nlines)
{
   osivx_iotag_t ctag;

   /* scan to header */
   if (osivx_fscan_totag (fi, OSIVX_TAG_HEADER) ||
       osivx_fread_tag (fi, &ctag))
     dump_die(DUMP_EFILE);

   /* print out header */
   if (!omit_headrs) {
      fprintf(fo, ARGUMENT_MSG);
   }

   /* proceed to thru header */
   while (osivx_tagcmp (ctag, OSIVX_GAT_HEADER))
   {
      switch (ctag)
      {
      case OSIVX_TAG_ARGINT:
         dump_arguments_integer(fi,fo);
         break;
      case OSIVX_TAG_ARGSTR:
         dump_arguments_string(fi,fo);
         break;
      default:
         dump_die(DUMP_EFILE);
      }

      if (osivx_fread_tag(fi, &ctag)) {
         dump_die(DUMP_EFILE);
      }
   }

   /* nice padding */
   if (!omit_nlines) {
      fprintf(fo, "\n");
   }
}

static void
dump_vectors_grid (
   osivx_file_t * fi,
   osivx_int * len,
   osivx_ioreal_t ** pos)
{
   osivx_ioint_t xnum, ynum;
   osivx_ioint_t xstart, ystart;
   osivx_ioint_t xspace, yspace;

   /* reset the file */
   if (osivx_fseek(fi, 0, SEEK_SET)) {
      dump_die(DUMP_EFILE);
   }

   /* collect grid info */
   {
      long mark;
      if (osivx_fscan_totag (fi, OSIVX_TAG_HEADER) ||
          osivx_ftell (fi, &mark) ||
          osivx_fgrab_argint (fi, OSIVX_KEY_STX, &xstart) ||
          osivx_fseek (fi, mark, SEEK_SET) ||
          osivx_fgrab_argint (fi, OSIVX_KEY_STY, &ystart) ||
          osivx_fseek (fi, mark, SEEK_SET) ||
          osivx_fgrab_argint (fi, OSIVX_KEY_NMX, &xnum) ||
          osivx_fseek (fi, mark, SEEK_SET) ||
          osivx_fgrab_argint (fi, OSIVX_KEY_NMY, &ynum) ||
          osivx_fseek (fi, mark, SEEK_SET) ||
          osivx_fgrab_argint (fi, OSIVX_KEY_SPX, &xspace) ||
          osivx_fseek (fi, mark, SEEK_SET) ||
          osivx_fgrab_argint (fi, OSIVX_KEY_SPY, &yspace) ||
          osivx_fseek(fi, 0, SEEK_SET))
        dump_die(DUMP_EFILE);
   }

   /* allocate memory */
   *len = 2 * xnum * ynum;
   *pos = (osivx_ioreal_t *) malloc (*len * sizeof (osivx_ioreal_t));
   if (*pos == NULL) dump_die(DUMP_EMEM);

   /* make grid */
   {
      osivx_ioreal_t *run = *pos;

      osivx_int i, j;
      for (i = 0; i < ynum; i++) {
         for (j = 0; j < xnum; j++) {
            *run++ = xstart + xspace * j;
            *run++ = ystart + yspace * i;
         }
      }
   }
}

static void
dump_vectors_read (
   osivx_file_t * fi,
   osivx_int frm,
   osivx_int * len,
   osivx_ioreal_t ** disp)
{
   if (osivx_fscan_totag (fi, OSIVX_TAG_BODY) ||
       osivx_fgrab_vectors (fi, frm, len, disp) ||
       osivx_fseek(fi, 0, SEEK_SET))
     dump_die(DUMP_EFRAME, frm);
}

static dump_fmtbox *
dump_vectors_format (
   const char *fmt,
   osivx_ioreal_t * xptr,
   osivx_ioreal_t * yptr,
   osivx_ioreal_t * uptr,
   osivx_ioreal_t * vptr)
{
   static char *flags = "#0+- ";
   static char *digits = "0123456789";

   const char *c0, *c1;
   dump_fmtbox *head = NULL;
   dump_fmtbox *tail = NULL;

   c1 = c0 = fmt;
   while (strlen (c0) > 0)
   {
      dump_fmtbox *box;

      /* allocate memory for element */
      box = (dump_fmtbox *) malloc( sizeof(dump_fmtbox) );
      if (box == NULL) dump_die(DUMP_EMEM);
      if (head == NULL) head = tail = box;

      /* find format character */
      c1 = strchr (c0, '%');

      /* skip to end of format */
      if (c1 == NULL)
      {
         size_t len;

         c1 = c0 + strlen (c0) - 1;

         /* push pointer */
         box->addr = NULL;

         /* push string */
         len = c1 - c0 + 1;

         box->fmt = (char *) malloc ((len + 1) * sizeof (char));
         if (box->fmt == NULL) dump_die(DUMP_EMEM);

         strncpy (box->fmt, c0, len);
         *(box->fmt + len) = '\0';
      } else {
         size_t len;
 
         /* After `%' */
         c1++;

         /* skip escaped format */
         if (*c1 == '%')
         {
            c1++;
            continue;
         }

         /* skip over format string */
         c1 += strspn (c1, flags);
         c1 += strspn (c1, digits);
         if (*c1 == '.')
           c1 += strspn (c1+1, digits)+1;

         /* push pointer */
         switch (*c1)
         {
           case 'x': box->addr = xptr; break;
           case 'y': box->addr = yptr; break;
           case 'u': box->addr = uptr; break;
           case 'v': box->addr = vptr; break;
           default: dump_die(DUMP_EFORMAT,fmt);
         }

         /* push string */
         len = c1 - c0 + 1;

         box->fmt = (char *) malloc ((len + 1) * sizeof (char));
         if (box->fmt == NULL) dump_die(DUMP_EMEM);

         strncpy (box->fmt, c0, len);
         *(box->fmt + len - 1) = 'f';
         *(box->fmt + len) = '\0';
      }

      /* push box */
      tail->next = box;
      box->next = NULL;
      tail = box;

      /* goto next */
      c0 = c1 + 1;
   }

   return head;
}

static void
dump_vectors_print(
   FILE * file,
   const char *fmt,
   osivx_int len,
   osivx_ioreal_t * pos,
   osivx_ioreal_t * disp)
{
   osivx_ioreal_t x, y;
   osivx_ioreal_t u, v;
   dump_fmtbox *list;
   osivx_int i;

   /* convert format string */
   list = dump_vectors_format(fmt, &x, &y, &u, &v);

   /* traverse file */
   for (i = 0; i < len / 2; i++) {
      dump_fmtbox *tmp;

      x = *pos++;  y = *pos++;
      u = *disp++; v = *disp++;

      tmp = list;
      while (tmp) {
          if (tmp->addr) {
            fprintf (file, tmp->fmt, *(tmp->addr));
          } else {
            fprintf (file, tmp->fmt);
          }
          tmp = tmp->next;
      }
   }

   /* cleanup format string */
   while (list) {
      dump_fmtbox *tmp;
      tmp = list->next;
      free (list->fmt);
      free (list);
      list = tmp;
   }
}

static void
dump_vectors(
   osivx_file_t * fi,
   FILE * fo,
   long frm[3],
   char *fmt,
   int omit_headrs,
   int omit_nlines,
   int omit_fbreak)
{
   osivx_ioreal_t *pos;
   osivx_ioreal_t *disp;
   osivx_int plen, dlen;
   osivx_int i;

   /* print out header */
   if (!omit_headrs) {
      fprintf(fo, VECTORS_MSG);
   }

   /* generate grid data */
   dump_vectors_grid(fi, &plen, &pos);

   /* print out each frame */
   if (frm[2] < frm[0]) dump_die(DUMP_EINTER);
   if (frm[1] < 0) dump_die(DUMP_EINTER);

   for (i = frm[0]; i <= frm[2]; i+=frm[1]) {
      /* read in vector data */
      dump_vectors_read(fi, i, &dlen, &disp);

      /* check for corrupt file */
      if (plen != dlen) dump_die(DUMP_EFILE);

      /* print out message */
      if (!omit_headrs) {
         fprintf(fo, FRAMES_MSG, (long)i);
      }

      /* print out vectors */
      dump_vectors_print(fo, fmt, plen, pos, disp);

      /* nice padding */
      if (i+frm[1]<=frm[2] && !omit_fbreak) {
         fprintf(fo, "\n");
      }

      free (disp);
   }

   /* nice padding */
   if (!omit_nlines) {
      fprintf(fo, "\n");
   }

   /* clean up */
   free (pos);
}

static void
dump_correlation_grid(
   osivx_file_t * fi,
   osivx_int* len,
   osivx_ioreal_t** disp)
{
   osivx_ioint_t xdisp, ydisp;
   osivx_ioreal_t* runner;
   osivx_int i, j;
   long file_mark;

   /* collect grid info */
   if (osivx_fscan_totag (fi, OSIVX_TAG_HEADER) ||
       osivx_ftell (fi, &file_mark) ||
       osivx_fgrab_argint (fi, OSIVX_KEY_MDX, &xdisp) ||
       osivx_fseek (fi, file_mark, SEEK_SET) ||
       osivx_fgrab_argint (fi, OSIVX_KEY_MDY, &ydisp) ||
       osivx_fseek(fi, 0, SEEK_SET))
      dump_die(DUMP_EFILE);

   /* allocate memory for arrays */
   *len = (2 * xdisp + 1) * (2 * ydisp + 1);
   *disp = (osivx_ioreal_t *) malloc (2 * *len * sizeof (osivx_ioreal_t));
   if (!*disp) dump_die(DUMP_EMEM);

   /* make grid */
   runner = *disp;
   for (i = -ydisp; i <= ydisp; i++) {
      for (j = -xdisp; j <= xdisp; j++) {
         *runner++ = j;
         *runner++ = i;
      }
   }
}

static void
dump_correlation_map(
   osivx_file_t * fi,
   osivx_int frm,
   osivx_int wdw,
   osivx_int* len,
   osivx_ioreal_t** corr)
{
   if (osivx_fscan_totag (fi, OSIVX_TAG_BODY) ||
       osivx_fgrab_map (fi, frm, wdw, len, corr) ||
       osivx_fseek(fi, 0, SEEK_SET))
     dump_die(DUMP_EMAPS);
}


static dump_fmtbox*
dump_correlation_format (
   const char *fmt,
   osivx_ioreal_t * uptr,
   osivx_ioreal_t * vptr,
   osivx_ioreal_t * eptr)
{
   static char *flags = "#0+- ";
   static char *digits = "0123456789";

   const char *c0, *c1;
   dump_fmtbox *head = NULL;
   dump_fmtbox *tail = NULL;

   c1 = c0 = fmt;
   while (strlen (c0) > 0)
   {
      dump_fmtbox *box;

      /* allocate memory for element */
      box = (dump_fmtbox *) malloc(sizeof(dump_fmtbox));
      if (!box) dump_die(DUMP_EMEM);
      if (!head) head = tail = box;

      /* find format character */
      c1 = strchr (c0, '%');

      /* skip to end of format */
      if (c1 == NULL) {
         size_t len;    

         c1 = c0 + strlen (c0) - 1;

         /* push pointer */
         box->addr = NULL;

         /* push string */
         len = c1 - c0 + 1;

         box->fmt = (char *) malloc ((len + 1) * sizeof (char));
         if (box->fmt == NULL) dump_die(DUMP_EMEM);

         strncpy (box->fmt, c0, len);
         *(box->fmt + len) = '\0';
      } else {
         size_t len;

         /* After `%' */
         c1++;

         /* skip escaped format */
         if (*c1 == '%') {
            c1++;
            continue;
         }

          /* skip over format string */
          c1 += strspn (c1, flags);
          c1 += strspn (c1, digits);
          if (*c1 == '.') {
            c1 += strspn (c1+1, digits)+1;
          }

          /* push pointer */
          switch (*c1)
          {
          case 'u': box->addr = uptr; break;
          case 'v': box->addr = vptr; break;
          case 'e': box->addr = eptr; break;
          default: dump_die(DUMP_EFORMAT,fmt);
          }

          /* push string */
	  len = c1 - c0 + 1;

          box->fmt = (char *) malloc ((len + 1) * sizeof (char));
          if (box->fmt == NULL) dump_die(DUMP_EMEM);

          strncpy (box->fmt, c0, len);
          *(box->fmt + len - 1) = 'f';
          *(box->fmt + len) = '\0';
      }

      /* push box */
      tail->next = box;
      box->next = NULL;
      tail = box;

      /* goto next */
      c0 = c1 + 1;
   }

   return head;
}

static void
dump_correlation_print(
   FILE* fo,
   const char* fmt,
   osivx_int len,
   osivx_ioreal_t* disp,
   osivx_ioreal_t* corr)
{
   osivx_ioreal_t u, v, e;
   dump_fmtbox *list;
   osivx_int i;

   /* convert format string */
   list = dump_correlation_format(fmt, &u, &v, &e);

   /* traverse file */
   for (i = 0; i < len; i++) {
      dump_fmtbox *tmp;

      u = *disp++;
      v = *disp++;
      e = *corr++;

      tmp = list;
      while (tmp) {
        if (tmp->addr) {
           fprintf (fo, tmp->fmt, *(tmp->addr));
        } else {
           fprintf (fo, tmp->fmt);
        }
        tmp = tmp->next;
     }
   }

   /* cleanup format string */
   while (list) {
      dump_fmtbox*tmp;
      tmp = list->next;
      free (list->fmt);
      free (list);
      list = tmp;
   }
}

static void
dump_correlation(
   osivx_file_t* fi,
   FILE* fo,
   long frm[2],
   osivx_int wdw,
   const char* fmt,
   int omit_headrs,
   int omit_nlines,
   int omit_fbreak)
{
   osivx_int gridlen;
   osivx_int corrlen;
   osivx_ioreal_t* grid;
   osivx_ioreal_t* corr;
   osivx_int i;

   /* print out header */
   if (!omit_headrs) {
      fprintf(fo, CORRELATION_MSG);
   }

   /* get grid and correlation map */
   dump_correlation_grid(fi, &gridlen, &grid);

   /* print out each frame */
   if (frm[2] < frm[0]) dump_die(DUMP_EINTER);
   if (frm[1] < 1) dump_die(DUMP_EINTER);

   for (i = frm[0]; i <= frm[2]; i+=frm[1]) {
      /* read in correlation data */
      dump_correlation_map(fi, i,
        wdw, &corrlen, &corr);

      /* check for sanity */
      if (gridlen != corrlen)
         dump_die(DUMP_EFILE);

      /* print out message */
      if (!omit_headrs) {
         fprintf(fo, FRAMES_MSG, (long)i);
      }
      /* dump to screen */
      dump_correlation_print(fo, fmt, 
        gridlen, grid, corr);

      /* nice padding */
      if (i+frm[1]<=frm[2] && !omit_fbreak) {
         fprintf(fo, "\n");
      }

      free(corr);
   }

   /* add nice padding */
   if (!omit_nlines) {
      fprintf(fo, "\n");
   }

   /* cleanup */
   free(grid);
}

static void
dump_print(
   dump_args* a)
{
   osivx_file_t* input;
   FILE* output;

   /* no standard input */
   if (!strcmp(a->input_fn, "-"))
     dump_die(DUMP_ESTDIN);

   /* open input file */
   input = osivx_fopen(a->input_fn, "r");
   if (input == NULL) {
     dump_die(DUMP_EFOPEN, a->input_fn);
   }

   /* open output file */
   if (strcmp(a->output_fn, "-")) {
      output = fopen(a->output_fn, "w");
   } else {
      output = stdout;
   }

   /* check success */
   if (output == NULL) {
      dump_die(DUMP_EFOUT, a->output_fn); 
   }

   /* print arguments */
   if (a->print_args) {
      dump_arguments(input, output,
         a->omit_headrs, a->omit_nlines);
   }

   /* print vectors */
   if (a->print_vecs) {
      dump_vectors(input, output,
         a->frame_range, a->vecs_fmt,
         a->omit_headrs, a->omit_nlines,
         a->omit_fbreak);
   }

   /* print correlation maps */
   if (a->print_maps) {
      dump_correlation(input, output,
         a->frame_range, a->wind_num,
         a->corr_fmt, a->omit_headrs,
         a->omit_nlines, a->omit_fbreak);
   }

   /* close input file */
   osivx_fclose(input);
}

int
main(int argc, char** argv)
{
   unsigned n,mask[DUMP_NOPTS];
   dump_args arglist;
   dump_info optinfo;
   dump_packet packet;

   /* install signals */
   dump_signal_mkhard();
   dump_signal_mksoft();

   /* null argument state */
   packet.mask = mask;
   memset(mask,0,sizeof(mask));
   packet.arglist = &arglist;
   packet.optinfo = &optinfo;

   /* check for usage request */
   if (argc == 1) {
     puts(dump_usage);
     exit(0);
   }

#ifdef DEBUG
   /* verify the options */
   if (zopts_verify(dump_long_opts,
         dump_long_nopts) != -1 ||
       zopts_verify(dump_shrt_opts,
         dump_shrt_nopts) != -1) {
      abort();
   }
#endif

   /* pull command options */
   n = zopts_parse(argc, argv,
     dump_long_opts, dump_long_nopts,
     dump_shrt_opts, dump_shrt_nopts,
     &dump_optcb, &packet);

   /* set input filename */
   if (argc-n != 1) {
     dump_die(DUMP_EARGV); 
   }

   arglist.input_fn = argv[n];

   /* grab information from file */
   dump_setinfo(&packet);

   /* set default when not found */
   dump_setdft(&packet);

   /* unescape the input */
   dump_unescape(arglist.vecs_fmt);
   dump_unescape(arglist.corr_fmt);

   /* check for sanity */
   if (arglist.print_vecs &&
       arglist.print_maps)
     dump_die(DUMP_EBOTH);

   if (arglist.frame_range[0] >
       arglist.frame_range[2] ||
       arglist.frame_range[1] < 1) {
     dump_die(DUMP_EFRANGE);
   }

   /* actually do work */
   dump_print(packet.arglist);

   return 0; 
}

