/* File: tune.c 
 * 
 * Copyright (c) 2003-2005 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */ 
 
#include "config.h"

#include <zargs.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <fftw3.h>


/* make fftw floating type */
typedef double fftw_real;

typedef struct
{
   unsigned*  mask;

   int        nrows;
   int        ncols;
   char*      out_file;
} tune_packet;

enum {
   TUNE_NUM_ROWS,   TUNE_NUM_COLS,
   TUNE_ROWS_COLS,  TUNE_OUT_FILE,
   TUNE_NOPTS
};

static const char* tune_usage =
"Usage: osiv_tune -r row -c column [ -o output ]\n\n"
"More information is available in the 'osiv_tune' manpage as well\n"
"as in the online documentation.";

static unsigned tune_long_nopts = 3;
static zopts_item tune_long_opts[] = {
   {"wind-size",    ZOPTS_WARG,  TUNE_ROWS_COLS},
   {"wind-xsize",   ZOPTS_WARG,  TUNE_NUM_COLS},
   {"wind-ysize",   ZOPTS_WARG,  TUNE_NUM_ROWS},
   {NULL, 0, 0}
};

static unsigned tune_shrt_nopts = 4;
static zopts_item tune_shrt_opts[] = {
   {"c",            ZOPTS_WARG,  TUNE_NUM_COLS},
   {"n",            ZOPTS_WARG,  TUNE_ROWS_COLS},
   {"o",            ZOPTS_WARG,  TUNE_OUT_FILE},
   {"r",            ZOPTS_WARG,  TUNE_NUM_ROWS},
   {NULL, 0, 0}
};

enum {
   TUNE_EMEM,    TUNE_EPBUFF,
   TUNE_EPLOST,  TUNE_EPWONT,
   TUNE_EPNEED,  TUNE_EMULTD,
   TUNE_ENINT,   TUNE_EIROWS,
   TUNE_EICOLS,  TUNE_EMROWS,
   TUNE_EMCOLS,  TUNE_EARGV, 
   TUNE_EOUTF
};

static const char*
tune_strerr(
   int id)
{
   const char* s;

   switch (id) {
   case TUNE_EMEM:
      s = "out of memory";
      break;
   case TUNE_EPBUFF:
      s = "argument to option `%s\' is too long";
      break;
   case TUNE_EPLOST:
      s = "found unrecognized option `%s\'";
      break;
   case TUNE_EPWONT:
      s = "option `%s\' doesn't take argument";
      break;
   case TUNE_EPNEED:
      s = "option `%s\' requires an argument";
      break;
   case TUNE_EMULTD:
      s = "option %s defined more than once";
      break;
   case TUNE_ENINT:
      s = "option %s expects integer but got `%s\'";
      break;
   case TUNE_EIROWS:
      s = "number of rows must be greater than one";
      break;
   case TUNE_EICOLS:
      s = "number of columns must be greater than one";
      break;
   case TUNE_EMROWS:
      s = "rows argument is required";
      break;
   case TUNE_EMCOLS:
      s = "columns argument is required";
      break;
   case TUNE_EARGV:
      s = "program doesn't take arguments";
      break;
   case TUNE_EOUTF:
      s = "unable to open output file `%s\'";
      break;
   default:
      s = NULL;
   }

   return s;
}

static void
tune_die(
   int id,...)
{
   va_list ap;
   const char *c;

   fprintf(stderr,"osiv_tune error: ");
   c = tune_strerr(id);

   va_start(ap,id);
   vfprintf(stderr, c, ap);
   va_end(ap);

   fprintf(stderr, "\n");
   exit(1);
}

static void
tune_setint(
   unsigned* preset,
   const char* nm,
   const char* val,
   int* out)
{
   char* endptr = NULL;

   if (*preset)
     tune_die(TUNE_EMULTD,nm);

   *out = strtol(val,&endptr,10);
   if (*endptr != '\0' || val==endptr)
     tune_die(TUNE_ENINT,nm,val);

   *preset = 1; 
} 

static void
tune_setstr(
   unsigned* preset,
   const char* nm,
   const char* val,
   char** out)
{
   if (*preset)
     tune_die(TUNE_EMULTD,nm);

   *out = strdup(val);
   if (*out == NULL)
     tune_die(TUNE_EMEM);

   *preset = 1;
}

static void
tune_opt_cb(
   long int evt,
   const char* var,
   const char* val,
   void* data)
{
   tune_packet* m = (tune_packet*)data; 

   switch (evt)
   {
   case TUNE_NUM_ROWS:
      tune_setint(&m->mask[TUNE_NUM_ROWS],
        var, val, &m->nrows);
      break;
   case TUNE_NUM_COLS:
      tune_setint(&m->mask[TUNE_NUM_COLS],
        var, val, &m->ncols);
      break;
   case TUNE_OUT_FILE:
      tune_setstr(&m->mask[TUNE_OUT_FILE],
        var, val, &m->out_file);
      break;
   case TUNE_ROWS_COLS:
      tune_setint(&m->mask[TUNE_NUM_ROWS],
        var, val, &m->nrows);
      tune_setint(&m->mask[TUNE_NUM_COLS],
        var, val, &m->ncols);
      break;
   default: break; 
   }

   switch (evt)
   {
   case ZOPTS_ENEED:
      tune_die(TUNE_EPNEED,var);
   case ZOPTS_EWONT:
      tune_die(TUNE_EPWONT,var);
   case ZOPTS_ELOST:
      tune_die(TUNE_EPLOST,var);
   case ZOPTS_EBUFF:
      tune_die(TUNE_EPBUFF,var);
   }
}

static void
tune_setdft(
   tune_packet* m)
{
   if (!m->mask[TUNE_NUM_ROWS]) {
      tune_die(TUNE_EMROWS);
   }
   if (!m->mask[TUNE_NUM_COLS]) {
      tune_die(TUNE_EMCOLS);
   }
   if (!m->mask[TUNE_OUT_FILE]) {
      m->out_file = strdup("-");
   }
}

 
void tune_run(
   FILE* out_fp, 
   long rows, 
   long cols) 
{ 
   fftw_real *ft; 
   fftw_plan planf;
   fftw_plan planb; 
   long len;
 
   /* make array */ 
   len = 2*rows*(cols/2+1);
   ft = (fftw_real*)fftw_malloc(len*sizeof(fftw_real)); 
   if (ft == NULL) {
      tune_die(TUNE_EMEM);
   }
 
   /* construct plans */ 
   planf = fftw_plan_dft_r2c_2d(rows, cols,
      ft, (fftw_complex*)ft, FFTW_EXHAUSTIVE); 
   planb = fftw_plan_dft_c2r_2d(rows, cols,
      (fftw_complex*)ft, ft, FFTW_EXHAUSTIVE); 
 
   /* write wisdom */ 
   fftw_export_wisdom_to_file(out_fp); 
   
   /* clean up */ 
   fftw_free(ft); 
   fftw_destroy_plan(planf); 
   fftw_destroy_plan(planb); 
} 
 
int main(int argc, char** argv) 
{ 
   unsigned n,mask[TUNE_NOPTS];
   tune_packet packet;
   FILE *out_fp; 

   /* null argument state */
   packet.mask = mask;
   memset(mask, 0, sizeof(mask));

   /* check for usage request */
   if (argc == 1) {
      puts(tune_usage);
      exit(0);
   }

#ifdef DEBUG
   /* verify option lists */
   if (zopts_verify(tune_long_opts,
         tune_long_nopts) != -1 ||
       zopts_verify(tune_shrt_opts,
         tune_shrt_nopts) != -1 ) {
      abort();
   }
#endif

   /* pull command options */
   n = zopts_parse(argc, argv,
     tune_long_opts, tune_long_nopts,
     tune_shrt_opts, tune_shrt_nopts,
     &tune_opt_cb, &packet);

   /* check for no trailing junk */
   if (argc - n > 0) {
     tune_die(TUNE_EARGV);
   }

   /* set defaults when not found */
   tune_setdft(&packet);

   /* open output file */ 
   if ( strcmp(packet.out_file, "-"))  {
      out_fp = fopen(packet.out_file, "w"); 
   } else {
      out_fp = stdout; 
   }

   /* check for valid file */ 
   if (out_fp == NULL)  {
      tune_die(TUNE_EOUTF, packet.out_file);
   }
 
   /* run fftw tuning */ 
   tune_run(out_fp, packet.nrows, packet.ncols);
 
   /* close output file */ 
   fclose(out_fp); 
 
   /* return */ 
   return 0; 
} 
