/* FILE: filt.c
 *
 * Copyright (c) 2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivs-devel.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>


/* The following is the maximum line
 * length for an input or output line
 */
#define LINE_BUFLEN 4096



/**********************************************************
 * The following function writes the data for processing
 * by a filter program.  If an error is encountered it
 * returns 1; otherwise, it returns 0 on success.
 **********************************************************/
static osivs_int
osivs_filt_write(
   FILE* fp,
   void* data)
{
   osivs_pts_t* pts = (osivs_pts_t*)data;
   int i;

   for (i = 0; i < pts->num; i++) {
      fprintf(fp, "%.4f  %.4f  %.4f\n",
         pts->x[i], pts->y[i], pts->z[i]);
   }

   return 0;
}



/**********************************************************
 * The following function parses a single mandatory
 * element from a data line. It returns -1 if this is
 * the last element on the line, -2 if the element is bad,
 * or the length of the parsed element
 **********************************************************/
static osivs_int
osivs_filt_readreq(
   const char* c,
   double* out)
{
   const char* r=c;
   char *eptr;
   double fn;

   /* scan leading space */
   c += strspn(c, " \t\n");

   /* check for end of line */
   if (*c == '\0') {
      return -1;
   }

   /* parse in the element */
   fn = strtod(c, &eptr);
   if (c==eptr || !isspace(*eptr)) {
      return -2;
   }

   /* return the result */
   *out = fn;
   return eptr - r;
}


/**********************************************************
 * The following function parses a single optional element
 * from a data line.  It returns -1 if this is the last
 * element on the line, -2 if the element is bad, or the
 * length of the parsed element
 **********************************************************/
static osivs_int
osivs_filt_readopt(
   const char* c,
   double* out)
{
   const char* r=c;
   char *eptr;
   double fn;

   /* scan leading space */
   c += strspn(c, " \t\n");

   /* check for skip mark */
   if (*c == '*') {
      return c - r + 1;
   }

   /* check for end of line */ 
   if (*c == '\0') {
      return -1;
   }

   /* parse in element */
   fn = strtod(c, &eptr);
   if (c==eptr || !isspace(*eptr)) {
      return -2;
   }

   /* return the result */
   *out = fn;
   return eptr - r;
}


/**********************************************************
 * The following function reads the data produced by a
 * filter program. If an error is encountered it returns
 * 1; it returns 0 on success.
 **********************************************************/
static osivs_int
osivs_filt_read(
   FILE* fp,
   void* data)
{
   osivs_pts_t* pts = (osivs_pts_t*)data;

   char buf[LINE_BUFLEN];
   int idx=0, n;

   /* read and process a single line */ 
   while (fgets(buf, LINE_BUFLEN, fp)) {
      char *c=buf;

      /* skip any leading space */
      c += strspn(c, " \t\n");

      /* skip comment lines */
      if (c[0] == '#') {
         continue;
      }

      /* skip blank lines */
      if (c[0] == '\0') {
         continue;
      }

      /* check for too many points */
      if (idx == pts->num) {
         return OSIVS_EFMORE;
      }

      /* set default values */
      pts->w[idx] = 0.0;
      pts->sz[idx] = 1.0;
      pts->bt1[idx] = 1.0;
      pts->bt2[idx] = 1.0;

      /* read in u number */
      n = osivs_filt_readreq(c, pts->u + idx);
      if (n < 0) return OSIVS_EFFMT;
      else c += n;

      /* read in v number */
      n = osivs_filt_readreq(c, pts->v + idx);
      if (n < 0) return OSIVS_EFFMT;
      else c += n;

      /* read in optional w number */
      n = osivs_filt_readopt(c, pts->w + idx);
      if (n == -2) return OSIVS_EFFMT;
      else if (n == -1) {idx++; continue;}
      else c += n;

      /* read in optional size number */
      n = osivs_filt_readopt(c, pts->sz + idx);
      if (n == -2) return OSIVS_EFFMT;
      else if (n == -1) {idx++; continue;}
      else c += n;

      /* read in optional brite number */
      n = osivs_filt_readopt(c, pts->bt1 + idx);
      if (n == -2) return OSIVS_EFFMT;
      else if (n == -1) {idx++; continue;}
      else c += n;

      /* read in optional brite number */
      n = osivs_filt_readopt(c, pts->bt2 + idx);
      if (n == -2) return OSIVS_EFFMT;
      else if (n == -1) {idx++; continue;}
      else c += n;

      /* check for trailing junk */
      c += strspn(c, " \t\n");
      if (*c != '\0') return OSIVS_EFFMT;

      /* goto to next item */
      idx++;
   }

   /* check for too few points */
   if (idx != pts->num) {
      return OSIVS_EFLESS;
   }

   /* all done, success */
   return 0;
}


/**********************************************************
 * The following function forks off a child process and
 * pushes position data in and pull displacement, size,
 * and brightness data out.
 **********************************************************/
static osivs_int
osivs_filt_run(
   const char* file,
   const char* argv[],
   osivs_filt_cb_t in_cb,
   osivs_filt_cb_t out_cb,
   void* cb_data)
{
   char in_fname[20];
   char out_fname[20];

   FILE* in_fp=NULL, *out_fp=NULL;
   int in_fd=-1, out_fd=-1;
   int retval=0, pid, p_stat;
   sigset_t cur_set, prev_set;

   /* block SIGINT during copy */
   sigemptyset(&cur_set);
   sigaddset(&cur_set, SIGINT);
   sigprocmask(SIG_BLOCK, &cur_set, &prev_set);

   /* initialize temporay names */
   strcpy(in_fname, "/tmp/osiv-XXXXXX");
   strcpy(out_fname, "/tmp/osiv-XXXXXX");

   /* open up the required files */
   in_fd = mkstemp(in_fname);
   if (in_fd == -1) {
      retval = OSIVS_ETEMP;
      goto finished;
   }

   out_fd = mkstemp(out_fname);
   if (out_fd == -1) {
      retval = OSIVS_ETEMP;
      goto finished;
   }

   /* wrap the fd with buffers */
   in_fp = fdopen(in_fd, "w");
   if (in_fp == NULL) {
      retval = OSIVS_ETEMP;
      goto finished;
   }

   /* write the input data */
   retval = (*in_cb)(in_fp, cb_data);
   if (retval != 0) {
      retval = OSIVS_ETEMP;
      goto finished;
   }

   /* closes input file */
   fclose(in_fp);
   in_fp = NULL;

   /* fork the child process */
   if ((pid = fork()) < 0) {
      retval = OSIVS_EFORK;
      goto finished;
   }

   /* start child process */
   if (pid == 0) {
      /* reopen input file */
      in_fd = open(in_fname, O_RDONLY);
      if (in_fd == -1) exit(1);

      /* reassign stdin */
      close(0);
      dup(in_fd);

      /* reassign stdout */
      close(1);
      dup(out_fd);

      /* exec to program */
      execvp(file, (char*const*)argv);

      /* shouldn't get here */
      exit(1);
   }

   /* wait for program to finish */
   wait(&p_stat);
   if (!WIFEXITED(p_stat) ||
       WEXITSTATUS(p_stat))
   {
      retval = OSIVS_EFILT;
      goto finished;
   }

   /* reset the fd (fork shares) */
   lseek(out_fd, 0, SEEK_SET);

   /* wrap the fd with buffers */
   out_fp = fdopen(out_fd, "r");
   if (out_fp == NULL) {
      retval = OSIVS_ETEMP;
      goto finished;
   }

   /* read the output data */
   retval = (*out_cb)(out_fp, cb_data);
   if (retval != 0) {
      goto finished;
   }

finished:
   /* restore prior sigmask */
   sigprocmask(SIG_SETMASK, &prev_set, NULL);

   /* close input file */
   if (in_fp != NULL) {
      fclose(in_fp);
      unlink(in_fname);
   } else if (in_fd != -1) {
      close(in_fd);
      unlink(in_fname);
   }

   /* close output file */
   if (out_fp != NULL) {
      fclose(out_fp);
      unlink(out_fname);
   } else if (out_fd != -1) {
      close(out_fd);
      unlink(out_fname);
   }

   return retval;
}


/**********************************************************
 * The following function runs the x,y,z data through a
 * data filter in order to get the u,v,w,sz,bt data
 **********************************************************/
osivs_int
osivs_filt(
   osivs_opts_t* p,
   osivs_pts_t* pts)
{
   const char* argv[4];
   int r;

   argv[0] = "/bin/sh";
   argv[1] = "-c";
   argv[2] = p->filt_src;
   argv[3] = NULL;

   r = osivs_filt_run(
      argv[0], argv,
      &osivs_filt_write,
      &osivs_filt_read,
      (void*)pts);

  return r;
}
 
