/* FILE: osivs-devel.h
 * 
 * Copyright (c) 2003-2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __OSIVS_DEVEL_H__
#define __OSIVS_DEVEL_H__

#include "osivs.h"
#include "config.h"

#include <stdlib.h>
#include <stdio.h>


/* These types must be kept consistent with
 * the values produced by the random number
 * generator
 */
typedef int32_t osivs_ccp_t;   /* coordinate */
typedef int32_t osivs_ccs_t;   /* subtract */
typedef int64_t osivs_ccm_t;   /* multiply */

#define OSIVS_RAND_MAX        2147483647
#define OSIVS_CCP_HI_BIT      30


/* If the mininum distance is too large it will
 * be impossible to find a unique spot.  We max
 * out after some number of iterations.
 */
#define OSIVS_MAX_TRIES 1000


/* The following struct is used to specify
 * a target position for quadtree searches
 */
typedef struct
{
   osivs_ccp_t x;
   osivs_ccp_t y;
   osivs_ccp_t z; 
} osivs_qtarg_t;


/* The following struct is used to hold the
 * quadtree of particle positions
 */

struct osivs_qtree_t_ {
   int is_parent;

   union {
      /* data for children */
      osivs_qtarg_t pt;

      /* data for parents */
     struct osivs_qtree_t_* ch[4]; 
   } u;
};

typedef struct osivs_qtree_t_ osivs_qtree_t;


/* The following type is used to provide a
 * callback when traversing a quadtree
 */
typedef int (*osivs_qtree_cb_t) (const osivs_qtree_t*, void*);


/* The following type is used to store 
 * particle positions and statistics
 */
typedef struct 
{
  osivs_real* x;    /* x position */
  osivs_real* y;    /* y position */
  osivs_real* z;    /* z position */
  osivs_real* u;    /* x displacement */
  osivs_real* v;    /* y displacement */
  osivs_real* w;    /* z displacement */
  osivs_real* sz;   /* particle size */
  osivs_real* bt1;  /* brightness (field 1) */
  osivs_real* bt2;  /* brightness (field 2) */
  osivs_int num;
} osivs_pts_t;


/* The following type stores options
 * following parsing of osivs_args_t
 */
typedef struct
{
  const char*   out_fnam0;
  const char*   out_fnam1;
  const char*   filt_src;
  const char*   backg_src;
  osivs_int     rand_seed;
  osivs_int     image_xsize;
  osivs_int     image_ysize;
  osivs_real    image_zsize;
  osivs_int     image_len;
  osivs_int     image_depth;
  osivs_real    image_xedge;
  osivs_real    image_yedge;
  osivs_real    image_zedge;
  osivs_real    dot_space;
  osivs_real    dot_density;
  osivs_real    dot_size_min;
  osivs_real    dot_size_max;
  osivs_real    dot_size_avg;
  osivs_real    dot_size_dev;
  osivs_real    dot_brite_min;
  osivs_real    dot_brite_max;
  osivs_real    dot_brite_avg;
  osivs_real    dot_brite_dev;
} osivs_opts_t;


/* alloca.c */
void* osivs_malloc(size_t);
void* osivs_calloc(size_t, size_t);
void osivs_free(void*);

/* draw.c */
osivs_int osivs_draw(osivs_opts_t*,osivs_pts_t*,
   osivs_real*, osivs_int);

/* filt.c */
typedef osivs_int (*osivs_filt_cb_t) (FILE*,  void*);
osivs_int osivs_filt(osivs_opts_t*, osivs_pts_t*);

/* genr.c */
osivs_int osivs_genr(osivs_opts_t*, osivs_pts_t*);

/* qtree.c */
void osivs_qtree_init(osivs_qtree_t*);
void osivs_qtree_free(osivs_qtree_t*);

osivs_qtree_t* osivs_qtree_add(osivs_qtree_t*,
   const osivs_qtarg_t*);

int osivs_qtree_invoke(const osivs_qtree_t*,
  const osivs_qtarg_t*, unsigned int,
  osivs_qtree_cb_t, void*);

void osivs_qtree_convert(const osivs_qtree_t*,
  const osivs_qtarg_t*, const osivs_qtarg_t*,
  osivs_pts_t*);

/* rand.c */
uint32_t osivs_rand();
float osivs_randn();
void osivs_srand(int32_t);
double osivs_exp (double);

/* scat.c */
osivs_int osivs_scat(osivs_opts_t*, osivs_pts_t*);
osivs_int osivs_disp(osivs_opts_t*, osivs_pts_t*);

/* write.c */
osivs_int osivs_write(osivs_opts_t*,osivs_real*,const char*);

#endif

