/* FILE: osivs.c
 *
 * Copyright (c) 2003-2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivs-devel.h"


/**********************************************************
 * The following function checks the passed arguments to
 * verify that they are tenable.  A zero is returned if
 * they are acceptable, and an error code is returned if
 * they are invalid.
 **********************************************************/
static osivs_int
osivs_checkargs(
   osivs_args_t* a)
{
   /* check the image sizes */
   if (a->image_xsize < 1) {
      return OSIVS_EXSIZE;
   }
   if (a->image_ysize < 1) {
      return OSIVS_EYSIZE; 
   }
   if (a->image_zsize < 1E-8) {
      return OSIVS_EZSIZE;
   }
   if (a->image_depth < 1) {
      return OSIVS_EIBITS;
   }
   if (a->image_depth > 16) {
      return OSIVS_EIBITS;
   }
   if (a->image_xedge < 0) {
      return OSIVS_EXEDGE;
   }
   if (a->image_yedge < 0) {
      return OSIVS_EYEDGE;
   }
   if (a->image_zedge < 0) {
      return OSIVS_EZEDGE;
   }

   /* check particles sizes */
   if (a->dot_space < 0) {
      return OSIVS_ESPACE;
   }
   if (a->dot_density < 0) {
      return OSIVS_EDENS;
   }
   if (a->dot_size_min < 0) {
      return OSIVS_ESMIN;
   }
   if (a->dot_size_max < 0) {
      return OSIVS_ESMAX; 
   }
   if (a->dot_size_min >= a->dot_size_max) {
      return OSIVS_ESCMP;
   }
   if (a->dot_size_avg <= 0) {
      return OSIVS_ESAVG;
   }
   if (a->dot_size_dev < 0) {
      return OSIVS_ESDEV;
   }
   if (a->dot_size_avg < a->dot_size_min ||
       a->dot_size_avg > a->dot_size_max) {
      return OSIVS_ESBND;
   }


   /* check brightness */
   if (a->dot_brite_min < 0) {
      return OSIVS_EBMIN;
   }
   if (a->dot_brite_max < 0) {
      return OSIVS_EBMAX;
   }
   if (a->dot_brite_min >= a->dot_brite_max) {
      return OSIVS_EBCMP;
   }
   if (a->dot_brite_avg <= 0) {
      return OSIVS_EBAVG;
   }
   if (a->dot_brite_dev < 0) {
      return OSIVS_EBDEV;
   }
   if (a->dot_brite_avg < a->dot_brite_min ||
       a->dot_brite_avg > a->dot_brite_max) {
      return OSIVS_EBBND;
   }

   return 0;
}



/**********************************************************
 * The following function uses the provided arguments to
 * generate a synthetic image.  This is the main function
 * which will be called from the library.  It returns a
 * zero on success, and an error code on failure.
 **********************************************************/
osivs_int osivs_run(osivs_args_t* a)
{
   osivs_real *im = NULL;
   osivs_int scode = 0;
   osivs_pts_t pts;
   osivs_opts_t p;

   /* check validity of input args */
   scode = osivs_checkargs(a);
   if (scode) goto finished;

   /* transfer args to internal struct */
   p.out_fnam0 = a->out_fnam0;
   p.out_fnam1 = a->out_fnam1;
   p.filt_src = a->filt_src;
   p.backg_src = a->backg_src;
   p.rand_seed = a->rand_seed;
   p.image_xsize = a->image_xsize;
   p.image_ysize = a->image_ysize;
   p.image_zsize = a->image_zsize;
   p.image_len = a->image_xsize * a->image_ysize;
   p.image_depth = a->image_depth;
   p.image_xedge = a->image_xedge;
   p.image_yedge = a->image_yedge;
   p.image_zedge = a->image_zedge;
   p.dot_space = a->dot_space;
   p.dot_density = a->dot_density;
   p.dot_size_min = a->dot_size_min;
   p.dot_size_max = a->dot_size_max;
   p.dot_size_avg = a->dot_size_avg;
   p.dot_size_dev = a->dot_size_dev;
   p.dot_brite_min = a->dot_brite_min;
   p.dot_brite_max = a->dot_brite_max;
   p.dot_brite_avg = a->dot_brite_avg;
   p.dot_brite_dev = a->dot_brite_dev;

   /* allocate memory for image */
   im = (osivs_real*)osivs_malloc(p.image_len*sizeof(osivs_real));
   if (im == NULL)  {
      scode = OSIVS_EMEM;
      goto finished;
   }

   /* initialize rand number generator */
   osivs_srand(p.rand_seed);

   /* generate points */
   scode = osivs_genr(&p, &pts);
   if (scode) goto finished;

   /* data filter points */
   scode = osivs_filt(&p, &pts);
   if (scode) goto finished;

   /* randomize size and bright */
   scode = osivs_scat(&p, &pts);
   if (scode) goto finished;

   /* make field A */
   osivs_draw(&p, &pts, im, 0);

   /* write field A */
   scode = osivs_write(&p, im ,p.out_fnam0);
   if (scode) goto finished;

   /* move points */
   scode = osivs_disp(&p,&pts);
   if (scode) goto finished;

   /* draw field B */
   osivs_draw(&p, &pts, im, 1);

   /* write field B */
   scode = osivs_write(&p,im,p.out_fnam1);
   if (scode) goto finished;

finished:
   if (im) {
      osivs_free(im);
   }

   return scode;
}

