/* FILE: write.c
 *
 * Copyright (c) 2003-2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivs-devel.h"
#include <tiffio.h>


/**********************************************************
 * This function writes the image contained in the matrix
 * 'im' to the file specified by the string 'fn' using
 * options contained in 'p'.
 **********************************************************/
osivs_int
osivs_write (
   osivs_opts_t* p,
   osivs_real* im,		
   const char* fn)
{
   int scode=0;
   TIFF* tif=NULL;
   void* buf=NULL;
   int nbits=0;
  
   /* figure out number of bits */
   if (p->image_depth <= 8) {
      nbits = 8;
   } else if (p->image_depth <= 16) {
      nbits = 16;
   } else {
      scode = OSIVS_EIBITS;
      goto finished;
   }

   /* allocate memory for scanlines */
   buf = osivs_malloc(p->image_xsize*(nbits/8));
   if (buf == NULL) {
      scode = OSIVS_EMEM;
      goto finished;
   }

   /* open TIFF file */
   tif = TIFFOpen(fn, "w");
   if (tif == NULL) {
      scode = OSIVS_EOUTF;
      goto finished;
   }

   /* set TIFF attribute fields */
   TIFFSetField(tif,TIFFTAG_PHOTOMETRIC, 1);
   TIFFSetField(tif,TIFFTAG_SAMPLESPERPIXEL, 1);
   TIFFSetField(tif,TIFFTAG_BITSPERSAMPLE, nbits);
   TIFFSetField(tif,TIFFTAG_IMAGEWIDTH, p->image_xsize);
   TIFFSetField(tif,TIFFTAG_IMAGELENGTH, p->image_ysize);
   TIFFSetField(tif,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);

   /* write scanlines to file buffer */
   switch (nbits) {
   case 8:
   {
      osivs_int i,j;
      uint8_t* ln=buf;

      for (i = 0; i < p->image_ysize; i++) {
         for (j = 0; j < p->image_xsize; j++) {
            *(ln+j) = *im++;
         }
         TIFFWriteScanline(tif,buf,i,0);
      }
   } break;
   case 16:
   {
      osivs_int i,j;
      uint16_t* ln=buf;

      for (i = 0; i < p->image_ysize; i++) {
         for (j = 0; j < p->image_xsize; j++) {
            *(ln+j) = *im++;
         }
         TIFFWriteScanline(tif,buf,i,0);
      }
   } break;
   case 32:
   {
      osivs_int i,j;
      uint32_t* ln=buf;

      for (i = 0; i < p->image_ysize; i++) {
         for (j = 0; j < p->image_xsize; j++) {
            *(ln+j) = *im++;
         }
         TIFFWriteScanline(tif,buf,i,0);
      }
   } break;
   default:
      scode = OSIVS_EIBITS;
      goto finished;
   }

finished:
   /* close file */
   if (buf != NULL) {
      free(buf);
   }
 
   if (tif != NULL) {
      TIFFClose(tif);
   }

   return scode;
}

