/* FILE: alg5.c 
 *
 * Copyright (c) 2002-2006 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"

#include <fftw3.h>

/**********************************************************
 * The following function traverses the passed arguments
 * and returns a non-zero error code if they are invalid.
 **********************************************************/
osivx_int
osivx_alg5_checkargs(osivx_args_t * a)
{
   osivx_int fft_xoffset = (a->wind_xsuper - a->wind_xsize)/2;
   osivx_int fft_yoffset = (a->wind_ysuper - a->wind_ysize)/2;

   /* check for grid extending over first image */
   if (a->grid_xstart - fft_xoffset - a->disp_xoffset < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart - fft_xoffset - a->disp_xoffset + a->wind_xsuper +
        (a->grid_xsize - 1) * a->grid_xspacing >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart - fft_yoffset - a->disp_yoffset < 0) {
      return OSIVX_EYGRID;
   }
   if (a->grid_ystart - fft_yoffset - a->disp_yoffset + a->wind_ysuper +
        (a->grid_ysize - 1) * a->grid_yspacing >= a->image_ysize) {
      return OSIVX_EYGRID;
   }


   /* check for grid extending over second image */
   if (a->grid_xstart < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart + a->wind_xsize + (a->grid_xsize - 1)
        * a->grid_xspacing >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart < 0) {
      return OSIVX_EYGRID;
   } 
   if (a->grid_ystart + a->wind_ysize + (a->grid_ysize - 1)
        * a->grid_yspacing >= a->image_ysize) {
      return OSIVX_EYGRID;
   }

   return 0;
}


/**********************************************************
 * The following function converts the integer pixel
 * values in the array 'im' into floating-point values
 * stored into the 'ft' array.  It cannot generate errors.
 **********************************************************/
void
osivx_alg5_copy_wind0(
   const osivx_imcp_t* im,
   osivx_int image_xsize,
   fftw_real* ft,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int wind_xsuper,
   osivx_int wind_ysuper,
   osivx_int fft_xoffset,
   osivx_int fft_yoffset,
   osivx_int fft_xsize)
{
   osivx_int i, j;

   /* offset index */
   ft += wind_ysuper * fft_xsize - 1;

   /* offset image */
   im -= fft_xoffset + fft_yoffset * image_xsize;

   /* traverse array */
   for (i = wind_ysuper - 1; i >= 0; i--) {
      const osivx_imcp_t *i0 = im + i * image_xsize;

      for (j = fft_xsize - wind_xsuper; j > 0; j--) {
         *ft-- = 0;
      }

      for (j = wind_xsuper - 1; j >= 0; j--) {
         *ft-- = *(i0 + j);
      }
   }
}


/**********************************************************
 * The following function converts the integer pixel
 * values in the array 'im' into floating-point values
 * stored into the 'ft' array.  It pads the data with
 * zeros so that it fits into the super window array.
 * It cannot generate errors.
 **********************************************************/
void
osivx_alg5_copy_wind1(
   const osivx_imcp_t* im,
   osivx_int image_xsize,
   fftw_real* ft,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int wind_xsuper,
   osivx_int wind_ysuper,
   osivx_int fft_xoffset,
   osivx_int fft_yoffset,
   osivx_int fft_xsize)
{
   osivx_int i, j;

   /* offset index */
   ft += wind_ysuper * fft_xsize - 1;

   /* traverse array */
   i = wind_ysuper - fft_yoffset - wind_ysize;
   for (; i > 0; i--) {
      for (j = fft_xsize; j > 0; j--) {
         *ft-- = 0;
      }
   }

   for (i = wind_ysize - 1; i >= 0; i--) {
      const osivx_imcp_t *i0 = im + i * image_xsize;

      j = fft_xsize - fft_xoffset - wind_xsize;
      for (; j > 0; j--) {
         *ft-- = 0;
      }

      for (j = wind_xsize - 1; j >= 0; j--) {
         *ft-- = *(i0 + j);
      }

      for (j = fft_xoffset; j > 0; j--) {
         *ft-- = 0;
      }
   }

   for (i = fft_yoffset; i > 0; i--) {
      for (j = fft_xsize; j > 0; j--) {
         *ft-- = 0;
      }
   }
}


/**********************************************************
 * The following function calculates the in-place
 * cross-correlation of two Fourier transformed
 * windows.  It cannot generate an error.
 **********************************************************/
void
osivx_alg5_getcorr(
   fftw_real* a,
   fftw_real* b,
   osivx_int len)
{
   osivx_int i;

   for (i = 0; i < len; i += 2) {
      register fftw_real tmp;

      tmp = *(a + i) * *(b + i) +
          *(a + i + 1) * *(b + i + 1);
      *(a + i + 1) = *(a + i + 1) * *(b + i) -
           *(a + i) * *(b + i + 1);
      *(a + i) = tmp;
   }
}


/**********************************************************
 * The following function converts a correlation map
 * that is stored in the ``Fourier-form'' into the
 * standard form.  It cannot generate an error.
 **********************************************************/
void
osivx_alg5_map2std(
   osivx_ioreal_t* dest,
   fftw_real* src,
   osivx_int disp_xmax,
   osivx_int disp_ymax,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int fft_xsize)
{
   osivx_int i, j;

   i = wind_ysize - disp_ymax;
   for (; i < wind_ysize; i++) {
      fftw_real *i0 = src + i * fft_xsize;

      j = wind_xsize - disp_xmax;
      for (; j < wind_xsize; j++) {
         *dest++ = *(i0 + j);
      }

      for (j = 0; j <= disp_xmax; j++) {
         *dest++ = *(i0 + j);
      }
   }

   for (i = 0; i <= disp_ymax; i++) {
      fftw_real *i0 = src + i * fft_xsize;

      j = wind_xsize - disp_xmax;
      for (; j < wind_xsize; j++) {
         *dest++ = *(i0 + j);
      }

      for (j = 0; j <= disp_xmax; j++) {
         *dest++ = *(i0 + j);
      }
   }
}


/**********************************************************
 * The following function is called to initialize the
 * algorithm 5 data.  It returns zero on success and a
 * non-zero number on error.
 **********************************************************/
static osivx_int
osivx_alg5_do_init(
   osivx_wind_alg_t* alg,
   const osivx_opts_t* p,
   const osivx_wind_alg_t* master)
{
   fftw_real* ft0, *ft1;
   osivx_int max_fft_len;
   osivx_int mode_wind_xsuper;
   osivx_int mode_wind_ysuper;

   alg->udata[0] = NULL;
   alg->udata[1] = NULL;
   alg->udata[2] = NULL;
   alg->udata[3] = NULL;
   alg->udata[4] = NULL;

   /* find the maximum size of window */
   {
      osivx_int max_wind_xsize, max_wind_ysize;
      osivx_int max_wind_xsuper, max_wind_ysuper;

      if (osivx_grid_get_max_wind_sizes(p->grid_data,
           &max_wind_xsize, &max_wind_ysize,
           &max_wind_xsuper, &max_wind_ysuper)) {
         return OSIVX_EGRID;
      }

      max_fft_len = max_wind_ysuper * (
        2*(max_wind_xsuper / 2 + 1));
   }

   /* find the most common window sizes */
   {
      osivx_int mode_wind_xsize, mode_wind_ysize;

      if (osivx_grid_get_mode_wind_sizes(p->grid_data,
           &mode_wind_xsize, &mode_wind_ysize,
           &mode_wind_xsuper, &mode_wind_ysuper)) {
         return OSIVX_EGRID;
      }
   }


   /* allocate memory for sub-window */
   alg->udata[0] = (void*) fftw_malloc(
      max_fft_len * sizeof(fftw_real));
   alg->udata[1] = (void*) fftw_malloc(
      max_fft_len * sizeof(fftw_real));
   if (alg->udata[0] == NULL ||
       alg->udata[1] == NULL) {
      return OSIVX_EMEM;
   }

   ft0 = (fftw_real*) alg->udata[0];
   ft1 = (fftw_real*) alg->udata[1];

   /* generate fftw plans */
   alg->udata[2] = (void*)fftw_plan_dft_r2c_2d(
      mode_wind_ysuper, mode_wind_xsuper, ft0,
      (fftw_complex*)ft0, FFTW_ESTIMATE);
   alg->udata[3] = (void*)fftw_plan_dft_r2c_2d(
      mode_wind_ysuper, mode_wind_xsuper, ft1,
      (fftw_complex*)ft1, FFTW_ESTIMATE);
   alg->udata[4] = (void*)fftw_plan_dft_c2r_2d(
      mode_wind_ysuper, mode_wind_xsuper,
      (fftw_complex*)ft1, ft1, FFTW_ESTIMATE);
   if (alg->udata[2] == NULL ||
       alg->udata[3] == NULL ||
       alg->udata[4] == NULL) {
      return OSIVX_EFFTPLAN;
   }

   /* store the plan sizes */
   alg->udata[5] = (void*)((long)mode_wind_xsuper);
   alg->udata[6] = (void*)((long)mode_wind_ysuper);

   /* quiet compiler warnings */
   master = NULL;

   return 0;
}


/**********************************************************
 * The following function calculates displacement
 * vectors for iter_num windows.  It writes the x
 * displacement to *std_vec, and the y displacement
 * to *(std_vec+1).  If std_map is non-null, then
 * the correlation map is copied to that memory.
 * It returns zero on success and a non-zero number
 * on error.
 **********************************************************/
static osivx_int
osivx_alg5_do_window(
   osivx_wind_alg_t* alg,
   const osivx_opts_t* p,
   osivx_grid_iter_t* iter,
   const osivx_imcp_t* im0,
   const osivx_imcp_t* im1,
   osivx_ioreal_t* std_vec,
   osivx_ioreal_t* std_map)
{
   osivx_int scode = 0;

   osivx_int wind_xpos, wind_ypos;
   osivx_int wind_xsize, wind_ysize;
   osivx_int wind_xsuper, wind_ysuper;
   osivx_int fft_xoffset, fft_yoffset;
   osivx_int fft_xsize;


   osivx_int ind, ix, iy;
   osivx_real vec[2];  

   fftw_real* ft0 = (fftw_real*)alg->udata[0];
   fftw_real* ft1 = (fftw_real*)alg->udata[1];

   osivx_int mode_wind_xsuper =
     (osivx_int)((long)alg->udata[5]);
   osivx_int mode_wind_ysuper =
     (osivx_int)((long)alg->udata[6]);

   fftw_plan forw0 = NULL;
   fftw_plan forw1 = NULL;
   fftw_plan back1 = NULL;

   /* find the window position */
   scode = osivx_grid_iter_get_info(iter,
      &wind_xpos, &wind_ypos,
      &wind_xsize, &wind_ysize,
      &wind_xsuper, &wind_ysuper);
   if (scode) goto finished;

   /* conjure fftw plans */
   if (wind_xsuper == mode_wind_xsuper &&
       wind_ysuper == mode_wind_ysuper) {
      forw0 = (fftw_plan)alg->udata[2];
      forw1 = (fftw_plan)alg->udata[3];
      back1 = (fftw_plan)alg->udata[4];
   } else {
      forw0 = (void*)fftw_plan_dft_r2c_2d(
        wind_ysuper, wind_xsuper, ft0,
        (fftw_complex*)ft0, FFTW_ESTIMATE);
      forw1 = fftw_plan_dft_r2c_2d(
        wind_ysuper, wind_xsuper, ft1,
        (fftw_complex*)ft1, FFTW_ESTIMATE);
      back1 = (void*)fftw_plan_dft_c2r_2d(
        wind_ysuper, wind_xsuper,
        (fftw_complex*)ft1, ft1, FFTW_ESTIMATE);
   }

   /* calculate fft window sizes */
   fft_xoffset = (wind_xsuper - wind_xsize)/2;
   fft_yoffset = (wind_ysuper - wind_ysize)/2;
   fft_xsize = 2 * (wind_xsuper / 2 + 1);

   /* offset the first image */
   im0 -= p->disp_xoffset + p->disp_yoffset * p->image_xsize;

   /* displace to window coordinates */
   im0 += wind_xpos + wind_ypos * p->image_xsize;
   im1 += wind_xpos + wind_ypos * p->image_xsize;

   /* copy over windows */
   osivx_alg5_copy_wind0(im0, 
     p->image_xsize, ft0,
     wind_xsize, wind_ysize,
     wind_xsuper, wind_ysuper,
     fft_xoffset, fft_yoffset,
     fft_xsize);
   osivx_alg5_copy_wind1(im1, 
     p->image_xsize, ft1,
     wind_xsize, wind_ysize,
     wind_xsuper, wind_ysuper,
     fft_xoffset, fft_yoffset,
     fft_xsize);

   /* calculate fft */
   fftw_execute(forw0);
   fftw_execute(forw1);

   /* perform an in-place correlation */
   osivx_alg5_getcorr(ft1, ft0, wind_ysuper * fft_xsize);

   /* calculate the inverse fft */
   fftw_execute(back1);

   /* find the maximum value */
   ind = osivx_find_max_fftw(ft1,
      wind_xsuper, wind_ysuper,
      p->disp_xmax, p->disp_ymax,
      fft_xsize);

   ix = ind % fft_xsize;
   iy = ind / fft_xsize;

   /* calculate actual displacement */
   if (ix < wind_xsuper/2 + 1) {
      *(vec) = ix + p->disp_xoffset;
   } else {
      *(vec) = ix - wind_xsuper + p->disp_xoffset;
   }
   if (iy < wind_ysuper/2 + 1) {
      *(vec + 1) = iy + p->disp_yoffset;
   } else {
      *(vec + 1) = iy - wind_ysuper + p->disp_yoffset;
   }

   /* perform interpolation */
   if ((ix < p->disp_xmax || ix > wind_xsuper - p->disp_xmax) &&
       (iy < p->disp_ymax || iy > wind_ysuper - p->disp_ymax)) {
       osivx_int lft, rht, up, dwn;

       if (ix != 0) {
          lft = -1;
       } else {
          lft = wind_xsuper - 1;
       }

       if (ix != wind_xsuper - 1) {
          rht = 1;
       } else {
          rht = -(wind_xsuper - 1);
       }

       if (iy != 0) {
          up = -fft_xsize;
       } else {
          up = (wind_ysuper - 1) * fft_xsize;
       }

       if (iy != wind_ysuper - 1) {
          dwn = fft_xsize;
       } else {
          dwn = -(wind_ysuper - 1) * fft_xsize;
       }

       osivx_interp_max_fftw(p->interp_data,
          ft1 + ind, lft, rht, up, dwn,
          vec, vec + 1);
   }

   /* copy over found displacement */
   *(std_vec) = *(vec);
   *(std_vec + 1) = *(vec + 1);

   /* copy over correlation map */
   if (std_map != NULL) {
      osivx_alg5_map2std(std_map, ft1,
        p->disp_xmax, p->disp_ymax,
        wind_xsuper, wind_ysuper,
        fft_xsize);
   }

finished:

   if (wind_xsuper != mode_wind_xsuper &&
       wind_ysuper != mode_wind_ysuper) {
      fftw_destroy_plan(forw0);
      fftw_destroy_plan(forw1);
      fftw_destroy_plan(back1);
   }

   return scode;
}


/**********************************************************
 * The following function frees any memory that was
 * generated by an algorithm 5 instance.  It cannot
 * generate an error.
 **********************************************************/
static void
osivx_alg5_do_free(
   osivx_wind_alg_t* alg)
{
   if (alg->udata[0]) {
      fftw_free(alg->udata[0]);
   }
   if (alg->udata[1]) {
      fftw_free(alg->udata[1]);
   }
   if (alg->udata[2]) {
      fftw_destroy_plan((fftw_plan)alg->udata[2]);
   }
   if (alg->udata[3]) {
      fftw_destroy_plan((fftw_plan)alg->udata[3]);
   }
   if (alg->udata[4]) {
      fftw_destroy_plan((fftw_plan)alg->udata[4]);
   }
}


/**********************************************************
 * The following function frees any memory that was
 * generated by all algorithm 5 instances.  It cannot
 * generate an error.
 **********************************************************/
static void
osivx_alg5_do_glbl_free(
   osivx_wind_alg_t* alg)
{
   fftw_cleanup();
}


/**********************************************************
 * The following function assigns all of the hooks
 * that will be needed for running alg5. It cannot
 * generate an error.
 **********************************************************/
void
osivx_alg5_make(
   osivx_wind_alg_t* alg)
{
   alg->do_glbl_init = NULL;
   alg->do_init = osivx_alg5_do_init;
   alg->do_window = osivx_alg5_do_window;
   alg->do_free = osivx_alg5_do_free;
   alg->do_glbl_free = osivx_alg5_do_glbl_free;
}

