/* FILE: alg6.c 
 *
 * Copyright (c) 2002-2006 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"

#include <fftw3.h>

/**********************************************************
 * The following function traverses the passed arguments
 * and returns a non-zero error code if they are invalid.
 **********************************************************/
osivx_int
osivx_alg6_checkargs(osivx_args_t * a)
{
   osivx_int fft_xoffset = (a->wind_xsuper - a->wind_xsize)/2;
   osivx_int fft_yoffset = (a->wind_ysuper - a->wind_ysize)/2;

   /* check for grid extending over first image */
   if (a->grid_xstart - fft_xoffset - a->disp_xoffset - a->disp_xmax < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart - fft_xoffset - a->disp_xoffset + a->disp_xmax +
        a->wind_xsuper + (a->grid_xsize - 1) * a->grid_xspacing
        >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart - fft_yoffset - a->disp_yoffset - a->disp_ymax < 0) {
      return OSIVX_EYGRID;
   }
   if (a->grid_ystart - fft_yoffset - a->disp_yoffset + a->disp_ymax +
        a->wind_ysuper + (a->grid_ysize - 1) * a->grid_yspacing
        >= a->image_ysize) {
      return OSIVX_EYGRID;
   }


   /* check for grid extending over second image */
   if (a->grid_xstart < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart + a->wind_xsize + (a->grid_xsize - 1)
        * a->grid_xspacing >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart < 0) {
      return OSIVX_EYGRID;
   } 
   if (a->grid_ystart + a->wind_ysize + (a->grid_ysize - 1)
        * a->grid_yspacing >= a->image_ysize) {
      return OSIVX_EYGRID;
   }

   return 0;
}


/**********************************************************
 * The following function converts the integer pixel
 * values in the array 'im' into floating-point values
 * stored into the 'ft' array.  It cannot generate errors.
 **********************************************************/
static void
osivx_alg6_copy_wind0(
   const osivx_imcp_t* im,
   osivx_int image_xsize,
   fftw_real* ft,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int wind_xsuper,
   osivx_int wind_ysuper,
   osivx_int fft_xoffset,
   osivx_int fft_yoffset,
   osivx_int fft_xsize)
{
   /* we just reuse the algorithm 5 routine */
   osivx_alg5_copy_wind0(im, image_xsize, ft,
    wind_xsize, wind_ysize, wind_xsuper,
    wind_ysuper, fft_xoffset, fft_yoffset,
    fft_xsize);
}


/**********************************************************
 * The following function converts the integer pixel
 * values in the array 'im' into floating-point values
 * stored into the 'ft' array.  It pads the data with
 * zeros so that it fits into the super window array.
 * It cannot generate errors.
 **********************************************************/
static void
osivx_alg6_copy_wind1(
   const osivx_imcp_t* im,
   osivx_int image_xsize,
   fftw_real* ft,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int wind_xsuper,
   osivx_int wind_ysuper,
   osivx_int fft_xoffset,
   osivx_int fft_yoffset,
   osivx_int fft_xsize)
{
   /* we just reuse the algorithm 5 routine */
   osivx_alg5_copy_wind1(im, image_xsize, ft,
    wind_xsize, wind_ysize, wind_xsuper,
    wind_ysuper, fft_xoffset, fft_yoffset,
    fft_xsize);
}


/**********************************************************
 * The following function calculates the in-place
 * cross-correlation of two Fourier transformed
 * windows.  It cannot generate an error.
 **********************************************************/
static void
osivx_alg6_getcorr(
   fftw_real* a,
   fftw_real* b,
   osivx_int len)
{
   /* we just reuse the algorithm 5 routine */
   osivx_alg5_getcorr(a, b, len);
}


/**********************************************************
 * The following function calculates the summed, squared
 * difference between windows for the required neigbors.
 * The regions of interest begin at im0 and im1, and are
 * assumed to have memory at both positive and negative
 * displacements sufficient for the given displacements.
 * It cannot generate an error.
 **********************************************************/
static void
osivx_alg6_getmap(
   const osivx_opts_t * p,
   const osivx_imcp_t * im0,
   const osivx_imcp_t * im1,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_immw_t * map,
   osivx_int * ind)
{
   osivx_int usage;
   osivx_int z, i, j;


   /* we only evaluate those points which will
    * be required by the particular interpolation
    * engine in use.
    */
   usage = osivx_interp_usage(p->interp_data);
   if (usage == OSIVX_INTERP_EIGHT) {
      /* evaluate all eight neighbors */

      /* set default minimum */
      *ind = z = 0;

      for (i = -1; i <= 1; i++) {
         const osivx_imcp_t* i0;

         i0 = im0 - i * p->image_xsize;

         for (j = -1; j <= 1; j++) {
            /* just reuse the algorithm 1 routine */
            osivx_alg1_sumwind(i0 - j, im1,
              p->image_xsize,
              wind_xsize, wind_ysize,
              map + z);

            if (*(map + z) <= *(map + *ind)) {
               *ind = z;
            }
 
            z++; 
         }
      }

   } else if (usage == OSIVX_INTERP_FOUR) {
      /* evaluate only four neighbors */

      /* set default minimum - must be
       * first evaluated point to avoid
       * access to uninit memory
       */
      *ind = z = 1;

      /* map position in first column,
       * x displacement in second,
       * y displacement in third
       */
      const osivx_int displ[5][3] = {
        {1, 0, -1},
        {3, -1, 0},
        {4, 0, 0},
        {5, +1, 0},
        {7, 0, +1}
      };

      for (i = 0; i < 5; i++) {
         const osivx_imcp_t* i0;

         z = displ[i][0];

         i0 = im0 - displ[i][2] * p->image_xsize
            - displ[i][1] ;

         /* just reuse the algorithm 1 routine */
         osivx_alg1_sumwind(i0, im1,
           p->image_xsize,
           wind_xsize, wind_ysize,
           map + z);

         if (*(map + z) <= *(map + *ind)) {
            *ind = z;
         }
      } 

   } else if (usage == OSIVX_INTERP_NONE) {
      /* do nothing - we set *ind just so
       * that caller isn't moving an uninit.
       * variable around, the interpolation
       * engine isn't allowed to access the
       * map value
       */

      *ind = 4;
   }
}


/**********************************************************
 * The following function converts a correlation map
 * that is stored in the ``Fourier-form'' into the
 * standard form.  It cannot generate an error.
 **********************************************************/
static void
osivx_alg6_map2std(
   osivx_ioreal_t* dest,
   fftw_real* src,
   osivx_int disp_xmax,
   osivx_int disp_ymax,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int fft_xsize)
{
   /* we just reuse the algorithm 5 routine */
   osivx_alg5_map2std(dest, src,
     disp_xmax, disp_ymax,
     wind_xsize, wind_ysize,
     fft_xsize);
}


/**********************************************************
 * The following function is called to initialize the
 * algorithm 5 data.  It returns zero on success and a
 * non-zero number on error.
 **********************************************************/
static osivx_int
osivx_alg6_do_init(
   osivx_wind_alg_t* alg,
   const osivx_opts_t* p,
   const osivx_wind_alg_t* master)
{
   fftw_real* ft0, *ft1;
   osivx_int max_fft_len;
   osivx_int mode_wind_xsuper;
   osivx_int mode_wind_ysuper;

   alg->udata[0] = NULL;
   alg->udata[1] = NULL;
   alg->udata[2] = NULL;
   alg->udata[3] = NULL;
   alg->udata[4] = NULL;

   /* find the maximum size of window */
   {
      osivx_int max_wind_xsize, max_wind_ysize;
      osivx_int max_wind_xsuper, max_wind_ysuper;

      if (osivx_grid_get_max_wind_sizes(p->grid_data,
           &max_wind_xsize, &max_wind_ysize,
           &max_wind_xsuper, &max_wind_ysuper)) {
         return OSIVX_EGRID;
      }

      max_fft_len = max_wind_ysuper * (
        2*(max_wind_xsuper / 2 + 1));
   }

   /* find the most common window sizes */
   {
      osivx_int mode_wind_xsize, mode_wind_ysize;

      if (osivx_grid_get_mode_wind_sizes(p->grid_data,
           &mode_wind_xsize, &mode_wind_ysize,
           &mode_wind_xsuper, &mode_wind_ysuper)) {
         return OSIVX_EGRID;
      }
   }


   /* allocate memory for sub-window */
   alg->udata[0] = (void*) fftw_malloc(
      max_fft_len * sizeof(fftw_real));
   alg->udata[1] = (void*) fftw_malloc(
      max_fft_len * sizeof(fftw_real));
   if (alg->udata[0] == NULL ||
       alg->udata[1] == NULL) {
      return OSIVX_EMEM;
   }

   ft0 = (fftw_real*)alg->udata[0];
   ft1 = (fftw_real*)alg->udata[1];

   /* generate fftw plans */
   alg->udata[2] = (void*)fftw_plan_dft_r2c_2d(
      mode_wind_ysuper, mode_wind_xsuper, ft0,
      (fftw_complex*)ft0, FFTW_ESTIMATE);
   alg->udata[3] = (void*)fftw_plan_dft_r2c_2d(
      mode_wind_ysuper, mode_wind_xsuper, ft1,
      (fftw_complex*)ft1, FFTW_ESTIMATE);
   alg->udata[4] = (void*)fftw_plan_dft_c2r_2d(
      mode_wind_ysuper, mode_wind_xsuper,
      (fftw_complex*)ft1, ft1, FFTW_ESTIMATE);
   if (alg->udata[2] == NULL ||
       alg->udata[3] == NULL ||
       alg->udata[4] == NULL) {
      return OSIVX_EFFTPLAN;
   }

   /* store the plan sizes */
   alg->udata[5] = (void*)((long)mode_wind_xsuper);
   alg->udata[6] = (void*)((long)mode_wind_ysuper);

   /* quiet compiler warnings */
   master = NULL;

   return 0;
}


/**********************************************************
 * The following function calculates displacement
 * vectors for iter_num windows.  It writes the x
 * displacement to *std_vec, and the y displacement
 * to *(std_vec+1).  If std_map is non-null, then
 * the correlation map is copied to that memory.
 * It returns zero on success and a non-zero number
 * on error.
 **********************************************************/
static osivx_int
osivx_alg6_do_window(
   osivx_wind_alg_t* alg,
   const osivx_opts_t* p,
   osivx_grid_iter_t* iter,
   const osivx_imcp_t* im0,
   const osivx_imcp_t* im1,
   osivx_ioreal_t* std_vec,
   osivx_ioreal_t* std_map)
{
   osivx_int scode = 0;

   osivx_int wind_xpos, wind_ypos;
   osivx_int wind_xsize, wind_ysize;
   osivx_int wind_xsuper, wind_ysuper;
   osivx_int fft_xoffset, fft_yoffset;
   osivx_int fft_xsize;

   osivx_immw_t mqd_map[9];
   osivx_int cur_disp_x, cur_disp_y;

   osivx_int ind, ix, iy;
   osivx_real vec[2];  

   fftw_real* ft0 = (fftw_real*)alg->udata[0];
   fftw_real* ft1 = (fftw_real*)alg->udata[1];

   osivx_int mode_wind_xsuper =
     (osivx_int)((long)alg->udata[5]);
   osivx_int mode_wind_ysuper =
     (osivx_int)((long)alg->udata[6]);

   fftw_plan forw0 = NULL;
   fftw_plan forw1 = NULL;
   fftw_plan back1 = NULL;

   /* find the window position */
   scode = osivx_grid_iter_get_info(iter,
      &wind_xpos, &wind_ypos,
      &wind_xsize, &wind_ysize,
      &wind_xsuper, &wind_ysuper);
   if (scode) goto finished;

   /* conjure fftw plans */
   if (wind_xsuper == mode_wind_xsuper &&
       wind_ysuper == mode_wind_ysuper) {
      forw0 = (fftw_plan)alg->udata[2];
      forw1 = (fftw_plan)alg->udata[3];
      back1 = (fftw_plan)alg->udata[4];
   } else {
      forw0 = (void*)fftw_plan_dft_r2c_2d(
        wind_ysuper, wind_xsuper, ft0,
        (fftw_complex*)ft0, FFTW_ESTIMATE);
      forw1 = fftw_plan_dft_r2c_2d(
        wind_ysuper, wind_xsuper, ft1,
        (fftw_complex*)ft1, FFTW_ESTIMATE);
      back1 = (void*)fftw_plan_dft_c2r_2d(
        wind_ysuper, wind_xsuper,
        (fftw_complex*)ft1, ft1, FFTW_ESTIMATE);
   }

   /* calculate fft window sizes */
   fft_xoffset = (wind_xsuper - wind_xsize)/2;
   fft_yoffset = (wind_ysuper - wind_ysize)/2;
   fft_xsize = 2 * (wind_xsuper / 2 + 1);

   /* offset the first image */
   im0 -= p->disp_xoffset + p->disp_yoffset * p->image_xsize;

   /* displace to window coordinates */
   im0 += wind_xpos + wind_ypos * p->image_xsize;
   im1 += wind_xpos + wind_ypos * p->image_xsize;

   /* copy over windows */
   osivx_alg6_copy_wind0(im0, 
     p->image_xsize, ft0,
     wind_xsize, wind_ysize,
     wind_xsuper, wind_ysuper,
     fft_xoffset, fft_yoffset,
     fft_xsize);
   osivx_alg6_copy_wind1(im1,
     p->image_xsize, ft1,
     wind_xsize, wind_ysize,
     wind_xsuper, wind_ysuper,
     fft_xoffset, fft_yoffset,
     fft_xsize);

   /* calculate fft */
   fftw_execute(forw0);
   fftw_execute(forw1);

   /* perform an in-place correlation */
   osivx_alg6_getcorr(ft1, ft0, wind_ysuper * fft_xsize);

   /* calculate the inverse fft */
   fftw_execute(back1);

   /* find the maximum value */
   ind = osivx_find_max_fftw(ft1,
      wind_xsuper, wind_ysuper,
      p->disp_xmax, p->disp_ymax,
      fft_xsize);

   ix = ind % fft_xsize;
   iy = ind / fft_xsize;

   /* calculate actual displacement */
   if (ix < wind_xsuper/2 + 1) {
      cur_disp_x = ix;
   } else {
      cur_disp_x = ix - wind_xsuper;
   }
   if (iy < wind_ysuper/2 + 1) {
      cur_disp_y = iy;
   } else {
      cur_disp_y = iy - wind_ysuper;
   }

   /* convert displacements to floating-point */
   *(vec) = cur_disp_x;
   *(vec + 1) = cur_disp_y;

   /* perform interpolation */
   if ((ix < p->disp_xmax || ix > wind_xsuper - p->disp_xmax) &&
       (iy < p->disp_ymax || iy > wind_ysuper - p->disp_ymax)) {
       const osivx_imcp_t* disp_im0;

       disp_im0 = im0 - cur_disp_y * p->image_xsize
          - cur_disp_x ;

       /* calculate new map using least squares */
       osivx_alg6_getmap(p, disp_im0, im1,
          wind_xsize, wind_ysize,
          mqd_map, &ind);

       if (ind == 4 /* mid-point */) {
          /* if MQD minimum matches up
           * with FFT map maximum
           */

          osivx_interp_min_immw(p->interp_data,
             mqd_map + ind, -1, 1, -3, +3,
             vec, vec + 1);
 
       } else {
          /* if MQD and FFT mismatch,
           * fallback and use FFT map
           */
          osivx_int lft, rht, up, dwn;

          *(vec) = cur_disp_x;
          *(vec + 1) = cur_disp_y;

          if (ix != 0) {
             lft = -1;
          } else {
             lft = wind_xsuper - 1;
          }

          if (ix != wind_xsuper - 1) {
             rht = 1;
          } else {
             rht = -(wind_xsuper - 1);
          }

          if (iy != 0) {
             up = -fft_xsize;
          } else {
             up = (wind_ysuper - 1) * fft_xsize;
          }

          if (iy != wind_ysuper - 1) {
             dwn = fft_xsize;
          } else {
             dwn = -(wind_ysuper - 1) * fft_xsize;
          }

          osivx_interp_max_fftw(p->interp_data,
             ft1 + ind, lft, rht, up, dwn,
             vec, vec + 1);
       }
   }

   /* compensate for global pre-shift */
   *(vec) += p->disp_xoffset;
   *(vec + 1) += p->disp_yoffset;

   /* copy over found displacement */
   *(std_vec) = *(vec);
   *(std_vec + 1) = *(vec + 1);

   /* copy over correlation map */
   if (std_map != NULL) {
      osivx_alg6_map2std(std_map, ft1,
        p->disp_xmax, p->disp_ymax,
        wind_xsize, wind_ysize,
        fft_xsize);
   }

finished:

   if (wind_xsuper != mode_wind_xsuper &&
       wind_ysuper != mode_wind_ysuper) {
      fftw_destroy_plan(forw0);
      fftw_destroy_plan(forw1);
      fftw_destroy_plan(back1);
   }

   return scode;
}


/**********************************************************
 * The following function frees any memory that was
 * generated by an algorithm 5 instance.  It cannot
 * generate an error.
 **********************************************************/
static void
osivx_alg6_do_free(
   osivx_wind_alg_t* alg)
{
   if (alg->udata[0]) {
      fftw_free(alg->udata[0]);
   }
   if (alg->udata[1]) {
      fftw_free(alg->udata[1]);
   }
   if (alg->udata[2]) {
      fftw_destroy_plan((fftw_plan)alg->udata[2]);
   }
   if (alg->udata[3]) {
      fftw_destroy_plan((fftw_plan)alg->udata[3]);
   }
   if (alg->udata[4]) {
      fftw_destroy_plan((fftw_plan)alg->udata[4]);
   }
}


/**********************************************************
 * The following function frees any memory that was
 * generated by all algorithm 5 instances.  It cannot
 * generate an error.
 **********************************************************/
static void
osivx_alg6_do_glbl_free(
   osivx_wind_alg_t* alg)
{
   fftw_cleanup();
}


/**********************************************************
 * The following function assigns all of the hooks
 * that will be needed for running alg6. It cannot
 * generate an error.
 **********************************************************/
void
osivx_alg6_make(
   osivx_wind_alg_t* alg)
{
   alg->do_glbl_init = NULL;
   alg->do_init = osivx_alg6_do_init;
   alg->do_window = osivx_alg6_do_window;
   alg->do_free = osivx_alg6_do_free;
   alg->do_glbl_free = osivx_alg6_do_glbl_free;
}


