/* FILE: file.c 
 *
 * Copyright (c) 2002-2006 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"

#include <string.h>

/* make file types */
enum {
   OSIV_FILE_READ,
   OSIV_FILE_WRITE,
   OSIV_FILE_APPEND
};


/**********************************************************
 * The following function opens an osivx_file_t for reading
 * using the provided file pointer.  It requires only a
 * relative forward file seek.  It returns a NULL pointer
 * on error.
 **********************************************************/
static osivx_file_t *
osivx_fsopen_read(FILE * fp)
{
   uint16_t val;
   osivx_file_t *file = NULL;
   osivx_int scode = 0;
   int cc;

   file = (osivx_file_t *) osivx_malloc(sizeof(osivx_file_t));
   if (file == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   file->mode = OSIV_FILE_READ;

   file->stream = fp;
   if (fp == NULL ||
       fgetc(fp) != 0x4a || /* Magic: JAS */ 
       fgetc(fp) != 0x41 ||
       fgetc(fp) != 0x53) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if (fread(&val, sizeof(uint16_t), 1, fp) != 1) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if ((file->swap_bytes = (val & 0xff00)) != 0) {
      file->swap_bits = (val & 0x8000);
   } else {
      file->swap_bits = (val & 0x0080);
   }

   if ((file->iotag_size = fgetc(fp)) == EOF ||
       (file->ioint_size = fgetc(fp)) == EOF ||
       (file->ioreal_size = fgetc(fp)) == EOF) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if (file->iotag_size != sizeof(osivx_iotag_t) ||
       file->ioint_size != sizeof(osivx_ioint_t) ||
       file->ioreal_size != sizeof(osivx_ioreal_t)) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if ((file->ver_first = fgetc(fp)) == EOF ||
       (file->ver_second = fgetc(fp)) == EOF ||
       (file->ver_third = fgetc(fp)) == EOF) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if ((cc = fgetc(fp)) == EOF || cc < 0) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if (fseek(fp, cc, SEEK_CUR)) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   if ((file->pre_len = ftell(fp)) == -1) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

finished:

   if (scode) {
      if (file != NULL) {
         osivx_free(file);
         file = NULL;
      }
   }

   return file;
}


/**********************************************************
 * The following function opens an osivx_file_t for writing 
 * using the provided file pointer.  It returns a NULL
 * pointer on error.
 **********************************************************/
static osivx_file_t *
osivx_fsopen_write(FILE * fp)
{
   uint16_t unity = 1;
   osivx_file_t *file = NULL;
   osivx_int scode = 0;

   file = (osivx_file_t*)osivx_malloc(sizeof(osivx_file_t));
   if (file == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   file->mode = OSIV_FILE_WRITE;

   file->stream = fp;
   if (fputc((char) 0x4a, fp) == EOF || /* Magic: JAS */
       fputc((char) 0x41, fp) == EOF ||
       fputc((char) 0x53, fp) == EOF ||
       fwrite(&unity, sizeof(uint16_t), 1, fp) != 1 ||
       fputc((char) sizeof(osivx_iotag_t), fp) == EOF ||
       fputc((char) sizeof(osivx_ioint_t), fp) == EOF ||
       fputc((char) sizeof(osivx_ioreal_t), fp) == EOF ||
       fputc((char) OSIVX_VER_FIRST, fp) == EOF ||
       fputc((char) OSIVX_VER_SECOND, fp) == EOF ||
       fputc((char) OSIVX_VER_THIRD, fp) == EOF ||
       fputc((char) 0, fp) == EOF) {
      scode = OSIVX_EFILEWRITE;
      goto finished;
   }

finished:

   if (scode) {
      if (file != NULL) {
         osivx_free(file);
         file = NULL;
      }
   }

   return file;
}


/**********************************************************
 * The following function opens an osivx_file_t for appending
 * using the provided file pointer.  It requires absolute
 * file seeking.  It returns a NULL pointer on error.  
 **********************************************************/
static osivx_file_t *
osivx_fsopen_append(FILE * fp)
{
   osivx_file_t *file = NULL;
   osivx_int scode = 0;

   file = osivx_fsopen_read(fp);
   if (file == NULL) {
      scode = OSIVX_EFILEWRITE;
      goto finished;
   }

   scode = osivx_fseek(file, 0, SEEK_END);
   if (scode) goto finished;
   
finished:

   if (scode) {
      if (file != NULL) {
         osivx_free(file);
         file = NULL;
      }
   }

   return file;
}


/**********************************************************
 * The following function opens an osivx_file_t using the
 * provided file pointer and mode.  See above for any
 * contraints on file capabilities.  It returns a NULL
 * pointer on error.  
 **********************************************************/
osivx_file_t *
osivx_fsopen(FILE * fp, const char *md)
{
   osivx_int type = -1;

   if (strchr(md, 'r')) {
      if (type == -1) {
         type = OSIV_FILE_READ;
      } else {
         return NULL;
      }
   }
   if (strchr(md, 'w')) {
      if (type == -1) {
         type = OSIV_FILE_WRITE;
      } else {
         return NULL;
      }
   }
   if (strchr(md, 'a')) {
      if (type == -1) {
         type = OSIV_FILE_APPEND;
      } else {
         return NULL;
      }
   }

   switch (type) {
   case OSIV_FILE_READ:
      return osivx_fsopen_read(fp);
   case OSIV_FILE_WRITE:
      return osivx_fsopen_write(fp);
   case OSIV_FILE_APPEND:
      return osivx_fsopen_append(fp);
   default:
      return NULL;
   }
}


/**********************************************************
 * The following function opens an osivx_file_t using the
 * provided file path and mode.  See above for any
 * contraints on file capabilities.  It returns a NULL
 * pointer on error.  
 **********************************************************/
osivx_file_t *
osivx_fopen(const char *fn, const char *md)
{
   osivx_int type = -1;

   if (strchr(md, 'r')) {
      if (type == -1) {
         type = OSIV_FILE_READ;
      } else {
         return NULL;
      }
   }
   if (strchr(md, 'w')) {
      if (type == -1) {
         type = OSIV_FILE_WRITE;
      } else {
         return NULL;
      }
   }
   if (strchr(md, 'a')) {
      if (type == -1) {
         type = OSIV_FILE_APPEND;
      } else {
         return NULL;
      }
   }

   switch (type) {
   case OSIV_FILE_READ: {
      FILE *fp;

      fp = fopen(fn, "rb");
      if (fp == NULL) {
         return NULL;
      }
      return osivx_fsopen(fp, "r");
   }
   case OSIV_FILE_WRITE: {
      FILE *fp;

      fp = fopen(fn, "wb");
      if (fp == NULL) {
         return NULL;
      }
      return osivx_fsopen(fp, "w");
   }
   case OSIV_FILE_APPEND: {
      FILE *fp;

      fp = fopen(fn, "a+b");
      if (fp == NULL) {
         return NULL;
      }

      if (fseek(fp, 0, SEEK_SET)) {
         return NULL;
      }
      return osivx_fsopen(fp, "a");
   }
   default:
      return NULL;
   }
}


/**********************************************************
 * The following function closes an osivx_file_t.  It
 * returns 0 on success and an error code on failure.
 **********************************************************/
osivx_int
osivx_fclose(osivx_file_t * file)
{
   osivx_int scode = 0;

   if (fclose(file->stream) == EOF) {
      if (file->mode == OSIV_FILE_READ) {
         scode = OSIVX_EFILEREAD;
      } else {
         scode = OSIVX_EFILEWRITE;
      }
      goto finished;
   }

finished:

   if (file != NULL) {
      osivx_free(file);
   }

   return scode;
}


/**********************************************************
 * The following function finds the file position.  It
 * returns 0 on success and an error code on failure.
 **********************************************************/
osivx_int
osivx_ftell(osivx_file_t * file, long *pos)
{
   *pos = ftell(file->stream);

   if (*pos == -1) {
      return OSIVX_EFILEPOS;
   }

   *pos -= file->pre_len;
   return 0;
}


/**********************************************************
 * The following function sets the file position.  It
 * returns 0 on success and an error code on failure.
 **********************************************************/
osivx_int
osivx_fseek(osivx_file_t * file, long offset, int whence)
{
   if (whence == SEEK_SET) {
      offset += file->pre_len;
   }

   if (fseek(file->stream, offset, whence) == -1) {
      return OSIVX_EFILEPOS;
   }

   return 0;
}

