/* FILE: generic.c 
 *
 * Copyright (c) 2006 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"


/**********************************************************
 * The following function runs an algorithm that is
 * generic at the window level.  In other words it 
 * goes through every frame and invokes the algorithm
 * routines for every window.  It returns zero on 
 * success and an error code if algorithm fails.
 **********************************************************/
osivx_int
osivx_generic_wind_alg(
   osivx_wind_alg_t* alg,
   osivx_opts_t* p)
{
   osivx_int i;
   osivx_int scode = 0;

   osivx_imcp_t *im0 = NULL;
   osivx_imcp_t *im1 = NULL;
   osivx_ioreal_t *std_vec = NULL;
   osivx_ioreal_t *std_map = NULL;

   osivx_int grid_len;
   grid_len = osivx_grid_get_len(p->grid_data);

   /* init the generic-algorithm data */
   if (alg->do_glbl_init != NULL) {
      scode = alg->do_glbl_init(alg, p);
      if (scode) goto finished;
   }

   scode = alg->do_init(alg, p, alg);
   if (scode) goto finished;

   /* allocate memory for images */
   im0 = (osivx_imcp_t *) osivx_malloc(
      p->image_len * sizeof(osivx_imcp_t));
   im1 = (osivx_imcp_t *) osivx_malloc(
      p->image_len * sizeof(osivx_imcp_t));
   if (im0 == NULL || im1 == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* allocate memory for vectors */
   std_vec = (osivx_ioreal_t *) osivx_malloc(
      2 * grid_len * sizeof(osivx_ioreal_t));
   if (std_vec == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   if (p->write_map) {
      std_map = (osivx_ioreal_t *) osivx_malloc(
         p->map_len * sizeof(osivx_ioreal_t));
      if (std_map == NULL) {
         scode = OSIVX_EMEM;
         goto finished;
      }
   } else {
      std_map = NULL;
   }

   /* open body tag */
   scode = osivx_fwrite_tag(p->vectors_file, OSIVX_TAG_BODY);
   if (scode) goto finished;

   /* traverse image sequence */
   for (i = p->movie_start; i <= p->movie_finish; i += p->movie_skip)
   {
      osivx_int j;
      osivx_grid_iter_t iter;

      /* read the ith image */
      scode = osivx_movie_read(p->movie_afile, i, im0);
      if (scode) goto finished;

      /* preprocess ith image */
      scode = osivx_preproc_exec(p->preproc_adata, i, im0);
      if (scode) goto finished;

      /* read the ith image partner */
      scode = osivx_movie_read(p->movie_bfile,
         i + p->movie_set, im1);
      if (scode) goto finished;
 
      /* preprocess ith image partner */ 
      scode = osivx_preproc_exec(p->preproc_bdata,
         i + p->movie_set, im1);
      if (scode) goto finished;

      /* open frame tag */
      scode = osivx_fwrite_tag(p->vectors_file, OSIVX_TAG_FRAME);
      if (scode) goto finished;

      /* write frame number */
      scode = osivx_fwrite_framenum(p->vectors_file, i);
      if (scode) goto finished;

      /* run through each of the windows */
      osivx_grid_begin(p->grid_data, &iter);

      for (j = 0; 1/*we test below*/; j++) { 
         osivx_int at_end;

         /* get the current window info */
         at_end = osivx_grid_iter_at_end(&iter);
         if (at_end) break;

         /* dispatch the window algorithm */
         alg->do_window(alg, p, &iter, im0, im1,
            std_vec + 2 * j, std_map); 

         /* write error map to file */
         if (p->write_map) {
            scode = osivx_fwrite_tag(p->vectors_file,
               OSIVX_TAG_MAPBOX);
            if (scode) goto finished;

            scode = osivx_fwrite_mapnum(p->vectors_file, j);
            if (scode) goto finished;

            scode = osivx_fwrite_map(p->vectors_file,
               p->map_len, std_map);
            if (scode) goto finished;

            scode = osivx_fwrite_tag(p->vectors_file,
               OSIVX_GAT_MAPBOX);
            if (scode) goto finished;
         }

         osivx_grid_iter_next(&iter);
      }

      /* dump vectors to file */
      scode = osivx_fwrite_vectors(p->vectors_file,
         2 * grid_len, std_vec);
      if (scode) goto finished;

      /* close frame tag */
      scode = osivx_fwrite_tag(p->vectors_file, OSIVX_GAT_FRAME);
      if (scode) goto finished;
   }

   /* close body tag */
   scode = osivx_fwrite_tag(p->vectors_file, OSIVX_GAT_BODY);
   if (scode) goto finished;

finished:
   /* free generic-algorithm data */
   alg->do_free(alg);
   if (alg->do_glbl_free != NULL) {
      alg->do_glbl_free(alg);
   }

   /* free memory */
   if (im0) osivx_free(im0);
   if (im1) osivx_free(im1);
   if (std_vec) osivx_free(std_vec);
   if (std_map) osivx_free(std_map);

   return scode;
}


