/* FILE: save.c
 *
 * Copyright (c) 2002 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivx-devel.h"

#include <string.h>


/**********************************************************
 * The following function writes raw data to the given
 * osivx_file_t.  It returns a zero on success and 
 * an error code on failure.
 **********************************************************/
static osivx_int
osivx_fwrite(
   osivx_file_t * fp,
   size_t size,
   size_t nmemb,
   const void *ptr)
{
   if(fwrite(ptr, size, nmemb, fp->stream) != nmemb) {
      return OSIVX_EFILEWRITE;
   }

   return 0;
}


/**********************************************************
 * The following function writes an opening tag to
 * the given file.  It returns a zero on success and 
 * an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_tag(
   osivx_file_t * fp,
   osivx_iotag_t tag)
{
   return osivx_fwrite(fp, sizeof(osivx_iotag_t), 1, &tag);
}


/**********************************************************
 * The following function writes an integer argument
 * block to the given file.  It returns a zero on success
 * and an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_argint(
  osivx_file_t * fp,
  osivx_ioint_t key,
  osivx_ioint_t val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   blen = 2 * sizeof(osivx_ioint_t);

   scode = osivx_fwrite_tag(fp, OSIVX_TAG_ARGINT);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &key);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &val);
   if (scode) return scode;
  
   scode = osivx_fwrite_tag(fp, OSIVX_GAT_ARGINT);
   if (scode) return scode;

   return scode;
}


/**********************************************************
 * The following function writes a string argument block
 * to the given file.  It returns a zero on success and
 * an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_argstr(
  osivx_file_t * fp,
  osivx_ioint_t key,
  const char *str)
{
   size_t len;
   osivx_ioint_t blen;
   osivx_int scode = 0;

   if (str == NULL) {
      return scode;
   }

   len = strlen(str);
   blen = sizeof(osivx_ioint_t) + len + 1;

   scode = osivx_fwrite_tag(fp, OSIVX_TAG_ARGSTR);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &key);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(char), len + 1, str);
   if (scode) return scode;

   scode = osivx_fwrite_tag(fp, OSIVX_GAT_ARGSTR);
   if (scode) return scode;

   return scode;
}


/**********************************************************
 * The following function writes a frame number block
 * to the given file.  It returns a zero on success and
 * an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_framenum(
   osivx_file_t * fp,
   osivx_ioint_t val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   blen = sizeof(osivx_ioint_t);

   scode = osivx_fwrite_tag(fp, OSIVX_TAG_FRAMENUM);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &val);
   if (scode) return scode;

   scode = osivx_fwrite_tag(fp, OSIVX_GAT_FRAMENUM);
   if (scode) return scode;

   return scode;
}


/**********************************************************
 * The following function writes a displacement vectors
 * block to the given file.  It returns a zero on success
 * and an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_vectors(
   osivx_file_t * fp,
   osivx_ioint_t len,
   osivx_ioreal_t * array)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   blen = sizeof(osivx_ioint_t) + len * sizeof(osivx_ioreal_t);

   scode = osivx_fwrite_tag(fp, OSIVX_TAG_VECTORS);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &len);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioreal_t), len, array);
   if (scode) return scode;

   scode = osivx_fwrite_tag(fp, OSIVX_GAT_VECTORS);
   if (scode) return scode;

   return scode;
}


/**********************************************************
 * The following function writes a map number block
 * to the given file.  It returns a zero on success and
 * an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_mapnum(
   osivx_file_t * fp,
   osivx_ioint_t val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   blen = sizeof(osivx_ioint_t);

   scode = osivx_fwrite_tag(fp, OSIVX_TAG_MAPNUM);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &val);
   if (scode) return scode;

   scode = osivx_fwrite_tag(fp, OSIVX_GAT_MAPNUM);
   if (scode) return scode;

   return scode;
}


/**********************************************************
 * The following function writes a correlation map
 * block to the given file.  It returns a zero on success
 * and an error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_map(
   osivx_file_t * fp,
   osivx_ioint_t len,
   osivx_ioreal_t * map)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   blen = sizeof(osivx_ioint_t) + len * sizeof(osivx_ioreal_t);

   scode = osivx_fwrite_tag(fp, OSIVX_TAG_MAP);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioint_t), 1, &len);
   if (scode) return scode;

   scode = osivx_fwrite(fp, sizeof(osivx_ioreal_t), len, map);
   if (scode) return scode;

   scode = osivx_fwrite_tag(fp, OSIVX_GAT_MAP);
   if (scode) return scode;

   return scode;
}


/**********************************************************
 * The following function writes a file header block to
 * the given file.  It returns a zero on success and an
 * error code on failure.
 **********************************************************/
osivx_int
osivx_fwrite_header(
   osivx_file_t * fp,
   osivx_args_t * input)
{
   osivx_int scode = 0;

   /* open header tag */
   scode = osivx_fwrite_tag(fp, OSIVX_TAG_HEADER);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_ALG, input->alg_xcorr);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_PPR, input->alg_pproc);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_ITP, input->alg_interp);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SZX, input->image_xsize);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SZY, input->image_ysize);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_STX, input->grid_xstart);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_STY, input->grid_ystart);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SPX, input->grid_xspacing);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SPY, input->grid_yspacing);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_NMX, input->grid_xsize);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_NMY, input->grid_ysize);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_OFX, input->disp_xoffset);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_OFY, input->disp_yoffset);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_MDX, input->disp_xmax);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_MDY, input->disp_ymax);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_ITX, input->disp_xiter);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_ITY, input->disp_yiter);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_WSX, input->wind_xsize);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_WSY, input->wind_ysize);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_WLX, input->wind_xsuper);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_WLY, input->wind_ysuper);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SFR, input->movie_start);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_EFR, input->movie_finish);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SKP, input->movie_skip);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_SET, input->movie_set);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_NIT, input->num_iter);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_NTH, input->num_thrd);
   if (scode) return scode;

   scode = osivx_fwrite_argint(fp, OSIVX_KEY_WMP, input->write_map);
   if (scode) return scode;

   scode = osivx_fwrite_argstr(fp, OSIVX_KEY_MVA, input->movie_aname);
   if (scode) return scode;

   scode = osivx_fwrite_argstr(fp, OSIVX_KEY_MVB, input->movie_bname);
   if (scode) return scode;

   scode = osivx_fwrite_argstr(fp, OSIVX_KEY_DTN, input->vectors_name);
   if (scode) return scode;

   scode = osivx_fwrite_tag(fp, OSIVX_GAT_HEADER);
   if (scode) return scode;

   return scode;
}

